# Copyright (c) Acconeer AB, 2023-2026
# All rights reserved

from __future__ import annotations

import logging
from enum import Enum, auto
from typing import Callable, Mapping, Optional

import attrs
import h5py
import numpy as np

from PySide6 import QtCore
from PySide6.QtWidgets import QLabel, QPushButton, QVBoxLayout

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121, opser
from acconeer.exptool._core.docstrings import get_attribute_docstring
from acconeer.exptool.a121._h5_utils import _create_h5_string_dataset
from acconeer.exptool.a121.algo._plugins import (
    A121BackendPluginBase,
    A121ViewPluginBase,
)
from acconeer.exptool.a121.algo._utils import get_distances_m
from acconeer.exptool.a121.algo.parking import (
    ObstructionProcessor,
    get_ground_config,
    get_pole_config,
)
from acconeer.exptool.app.new import (
    AppModel,
    AttrsConfigEditor,
    BackendLogger,
    GeneralMessage,
    GroupBox,
    HandledException,
    Message,
    MiscErrorView,
    PgPlotPlugin,
    PidgetGroupFactoryMapping,
    PluginFamily,
    PluginGeneration,
    PluginPresetBase,
    PluginSpecBase,
    PluginState,
    PluginStateMessage,
    backend,
    icons,
    is_task,
    pidgets,
    visual_policies,
)

from ._processors import MAX_AMPLITUDE
from ._ref_app import (
    DetailedStatus,
    RefApp,
    RefAppConfig,
    RefAppContext,
    RefAppResult,
    _load_algo_data,
    get_sensor_configs,
)


log = logging.getLogger(__name__)


@attrs.mutable(kw_only=True)
class SharedState:
    sensor_id: int = attrs.field(default=1)
    config: RefAppConfig = attrs.field(factory=get_ground_config)
    context: RefAppContext = attrs.field(default=None)


class PluginPresetId(Enum):
    GROUND = auto()
    POLE = auto()


@attrs.frozen(kw_only=True)
class SetupMessage(GeneralMessage):
    ref_app_config: RefAppConfig
    session_config: a121.SessionConfig
    sensor_id: int
    metadata: a121.Metadata
    name: str = attrs.field(default="setup", init=False)
    recipient: backend.RecipientLiteral = attrs.field(default="plot_plugin", init=False)


class BackendPlugin(A121BackendPluginBase[SharedState]):
    PLUGIN_PRESETS: Mapping[int, Callable[[], RefAppConfig]] = {
        PluginPresetId.GROUND.value: lambda: get_ground_config(),
        PluginPresetId.POLE.value: lambda: get_pole_config(),
    }

    def __init__(
        self,
        callback: Callable[[Message], None],
        generation: PluginGeneration,
        key: str,
    ) -> None:
        super().__init__(callback=callback, generation=generation, key=key)

        self._recorder: Optional[a121.H5Recorder] = None
        self._ref_app_instance: Optional[RefApp] = None
        self._log = BackendLogger.getLogger(__name__)

        self.restore_defaults()

    def _load_from_cache(self, file: h5py.File) -> None:
        self.shared_state.config = RefAppConfig.from_json(file["config"][()])

        context = opser.try_deserialize(file["context"], RefAppContext)
        if context is None:
            self.send_status_message(
                "Could not load cached context. Falling back to empty context"
            )
            context = RefAppContext()

        self.shared_state.context = context

    @is_task
    def restore_defaults(self) -> None:
        self.shared_state = SharedState()
        self.broadcast()

    @is_task
    def update_sensor_id(self, *, sensor_id: int) -> None:
        self.shared_state.sensor_id = sensor_id
        self.broadcast()

    def _sync_sensor_ids(self) -> None:
        if self.client is not None:
            sensor_ids = self.client.server_info.connected_sensors

            if len(sensor_ids) > 0 and self.shared_state.sensor_id not in sensor_ids:
                self.shared_state.sensor_id = sensor_ids[0]

    @is_task
    def update_config(self, *, config: RefAppConfig) -> None:
        self.shared_state.config = config
        self.broadcast()

    def save_to_cache(self, file: h5py.File) -> None:
        _create_h5_string_dataset(file, "config", self.shared_state.config.to_json())
        opser.serialize(self.shared_state.context, file.create_group("context"))

    @is_task
    def set_preset(self, preset_id: int) -> None:
        preset_config = self.PLUGIN_PRESETS[preset_id]
        self.shared_state.config = preset_config()
        self.broadcast()

    def load_from_record_setup(self, *, record: a121.H5Record) -> None:
        algo_group = record.get_algo_group(self.key)
        sensor_id, config, context = _load_algo_data(algo_group)
        self.shared_state.config = config
        self.shared_state.sensor_id = sensor_id
        self.shared_state.context = context

    def _start_session(self, recorder: Optional[a121.H5Recorder]) -> None:
        assert self.client
        self._ref_app_instance = RefApp(
            client=self.client,
            sensor_id=self.shared_state.sensor_id,
            ref_app_config=self.shared_state.config,
            context=self.shared_state.context,
        )
        self._ref_app_instance.start(recorder)

        metadata = self.client.extended_metadata[0][self.shared_state.sensor_id]
        self.callback(
            SetupMessage(
                ref_app_config=self.shared_state.config,
                metadata=metadata,
                sensor_id=self.shared_state.sensor_id,
                session_config=self.client.session_config,
            )
        )

    def end_session(self) -> None:
        if self._ref_app_instance is None:
            raise RuntimeError
        self._ref_app_instance.stop()

    def get_next(self) -> None:
        assert self.client
        if self._ref_app_instance is None:
            raise RuntimeError
        result = self._ref_app_instance.get_next()

        self.callback(backend.PlotMessage(result=result))

    @is_task
    def calibrate_detector(self) -> None:
        if self._started:
            raise RuntimeError

        if self.client is None:
            raise RuntimeError

        if not self.client.connected:
            raise RuntimeError

        self.callback(PluginStateMessage(state=PluginState.LOADED_BUSY))

        try:
            self._ref_app_instance = RefApp(
                client=self.client,
                sensor_id=self.shared_state.sensor_id,
                ref_app_config=self.shared_state.config,
                context=RefAppContext(),
            )
            self._ref_app_instance.calibrate_ref_app()
        except Exception as exc:
            msg = "Failed to calibrate detector"
            raise HandledException(msg) from exc
        finally:
            self.callback(PluginStateMessage(state=PluginState.LOADED_IDLE))

        self.shared_state.context = self._ref_app_instance.context
        self.broadcast()


class PlotPlugin(PgPlotPlugin):
    def __init__(self, app_model: AppModel) -> None:
        super().__init__(app_model=app_model)
        self._plot_job: Optional[RefAppResult] = None
        self._is_setup = False
        self.obstruction_text_timeout = 0

    def handle_message(self, message: backend.GeneralMessage) -> None:
        if isinstance(message, backend.PlotMessage):
            self._plot_job = message.result
        elif isinstance(message, SetupMessage):
            self.setup(
                message.ref_app_config,
                message.metadata,
                message.sensor_id,
                message.session_config,
            )
            self._is_setup = True
        else:
            log.warn(f"{self.__class__.__name__} got an unsupported command: {message.name!r}.")

    def draw(self) -> None:
        if not self._is_setup or self._plot_job is None:
            return
        try:
            self.draw_plot_job(ref_app_result=self._plot_job)
        finally:
            self._plot_job = None

    def setup(
        self,
        ref_app_config: RefAppConfig,
        metadata: a121.Metadata,
        sensor_id: int,
        session_config: a121.SessionConfig,
    ) -> None:
        self.plot_layout.clear()

        self.metadata = metadata
        self.ref_app_config = ref_app_config
        sensor_configs = get_sensor_configs(ref_app_config)
        display_config = sensor_configs["base_config"]
        self.sensor_id = sensor_id
        self.session_config = session_config
        self.sensor_config = display_config
        self.distances = get_distances_m(display_config, metadata)

        if self.ref_app_config.obstruction_detection:
            obstruction_config = sensor_configs["obstruction_config"]
            self.obs_distances = get_distances_m(obstruction_config, metadata)
            self.obs_x_thres, self.obs_y_thres = ObstructionProcessor.get_thresholds(
                ref_app_config.obstruction_distance_threshold, self.obs_distances
            )

        win = self.plot_layout
        # Define pens and font.
        blue_color = et.utils.color_cycler(0)
        orange_color = et.utils.color_cycler(1)
        brush = et.utils.pg_brush_cycler(0)

        self.blue = dict(
            pen=pg.mkPen(blue_color, width=2),
            symbol="o",
            symbolSize=1,
            symbolBrush=brush,
            symbolPen="k",
        )
        self.orange = dict(
            pen=pg.mkPen(orange_color, width=2),
            symbol="o",
            symbolSize=1,
            symbolBrush=brush,
            symbolPen="k",
        )
        self.blue_transparent_pen = pg.mkPen(f"{blue_color}50", width=2)
        self.orange_transparent_pen = pg.mkPen(f"{orange_color}50", width=2)

        brush_dot = et.utils.pg_brush_cycler(1)

        # Signature plot.
        self.sig_plot = win.addPlot(row=0, col=0, colspan=2)
        self.sig_plot.setTitle("Sampled Signatures")
        self.sig_plot.setMenuEnabled(False)
        self.sig_plot.showGrid(x=True, y=True)
        self.sig_plot.addLegend()
        self.sig_plot.setLabel("left", "Normalized energy")
        self.sig_plot.setLabel("bottom", "Distance (m)")
        self.sig_plot.addItem(pg.PlotDataItem())
        self.sig_plot_x_range = (
            min(self.distances),
            max(self.distances) + ref_app_config.weighted_distance_threshold_m,
        )
        self.sig_plot.setXRange(self.sig_plot_x_range[0], self.sig_plot_x_range[1])
        self.sig_plot.setYRange(0, 100)
        symbol_kw_main = dict(
            symbol="o", symbolSize=7, symbolBrush=brush, symbolPen=None, pen=None
        )
        self.sig_plot_curve = self.sig_plot.plot(**symbol_kw_main)
        energy_threshold_line = pg.InfiniteLine(
            angle=0, pen=pg.mkPen("k", width=1.5, style=QtCore.Qt.PenStyle.DashLine)
        )
        energy_threshold_line.setVisible(True)
        energy_threshold_line.setPos(ref_app_config.amplitude_threshold)
        self.sig_plot.addItem(energy_threshold_line)

        self.sig_plot_cluster_start = pg.InfiniteLine(angle=90, pen=pg.mkPen("k", width=1.5))
        self.sig_plot_cluster_start.setVisible(True)
        self.sig_plot.addItem(self.sig_plot_cluster_start)

        self.sig_plot_cluster_end = pg.InfiniteLine(angle=90, pen=pg.mkPen("k", width=1.5))
        self.sig_plot_cluster_end.setVisible(True)
        self.sig_plot.addItem(self.sig_plot_cluster_end)

        self.sig_plot_smooth_max = et.utils.SmoothMax(self.session_config.update_rate)

        parking_car_html = (
            '<div style="text-align: center">'
            '<span style="color: #FFFFFF;font-size:15pt;">'
            "{}</span></div>".format("Parked car detected!")
        )
        self.parking_car_text_item = pg.TextItem(
            html=parking_car_html,
            fill=orange_color,
            anchor=(0.5, 0),
        )
        parking_no_car_html = (
            '<div style="text-align: center">'
            '<span style="color: #FFFFFF;font-size:15pt;">'
            "{}</span></div>".format("No car detected.")
        )
        self.parking_no_car_text_item = pg.TextItem(
            html=parking_no_car_html,
            fill=blue_color,
            anchor=(0.5, 0),
        )

        self.sig_plot.addItem(self.parking_car_text_item)
        self.sig_plot.addItem(self.parking_no_car_text_item)
        self.parking_car_text_item.hide()
        self.parking_no_car_text_item.hide()

        self.cluster_width = ref_app_config.weighted_distance_threshold_m

        # Obstruction plot.
        if self.ref_app_config.obstruction_detection:
            self.obstruction_plot = win.addPlot(row=1, col=1)
            self.obstruction_plot.setTitle("Obstruction Detection Signatures")
            self.obstruction_plot.setMenuEnabled(False)
            self.obstruction_plot.showGrid(x=True, y=True)
            self.obstruction_plot.addLegend()
            self.obstruction_plot.setLabel("left", "Average energy")
            self.obstruction_plot.setLabel("bottom", "Distance (m)")
            self.obstruction_plot.addItem(pg.PlotDataItem())
            self.obstruction_plot.setXRange(min(self.obs_distances), max(self.obs_distances))
            self.obstruction_plot.setYRange(0, MAX_AMPLITUDE)  # Set to standard
            self.obstruction_plot_curve = self.obstruction_plot.plot(**self.orange)

            symbol_obstruction_dot = dict(
                symbol="o",
                symbolSize=7,
                symbolBrush=brush_dot,
                symbolPen=None,
                pen=None,
            )
            self.obstruction_plot_point = self.obstruction_plot.plot(**symbol_obstruction_dot)

            symbol_kw_main = dict(
                symbol="o", symbolSize=7, symbolBrush=brush, symbolPen=None, pen=None
            )
            self.obstruction_plot_center = self.obstruction_plot.plot(**symbol_kw_main)

            self.obstruction_center_rect = pg.QtWidgets.QGraphicsRectItem(0, 0, 0.01, 0.01)
            self.obstruction_center_rect.setPen(self.orange_transparent_pen)
            self.obstruction_plot.addItem(self.obstruction_center_rect)

            obstruction_html = (
                '<div style="text-align: center">'
                '<span style="color: #FFFFFF;font-size:15pt;">'
                "{}</span></div>".format("Obstruction detected!")
            )
            self.obstruction_text_item = pg.TextItem(
                html=obstruction_html,
                fill=orange_color,
                anchor=(0.5, 0),
            )
            no_obstruction_html = (
                '<div style="text-align: center">'
                '<span style="color: #FFFFFF;font-size:15pt;">'
                "{}</span></div>".format("No obstruction detected.")
            )
            self.no_obstruction_text_item = pg.TextItem(
                html=no_obstruction_html,
                fill=blue_color,
                anchor=(0.5, 0),
            )

            obs_text_x_pos = (
                min(self.obs_distances) + (max(self.obs_distances) - min(self.obs_distances)) * 0.5
            )
            obs_text_y_pos = MAX_AMPLITUDE * 0.9

            self.obstruction_text_item.setPos(obs_text_x_pos, obs_text_y_pos)
            self.no_obstruction_text_item.setPos(obs_text_x_pos, obs_text_y_pos)

            self.obstruction_plot.addItem(self.obstruction_text_item)
            self.obstruction_plot.addItem(self.no_obstruction_text_item)
            self.obstruction_text_item.hide()
            self.no_obstruction_text_item.hide()

        # Parking info plot.
        self.parking_plot = win.addPlot(row=1, col=0)
        self.parking_plot.setTitle("Noise adjusted amplitude")
        self.parking_plot.setMenuEnabled(False)
        self.parking_plot.showGrid(x=True, y=True)
        self.parking_plot.setLabel("left", "Normalized energy")
        self.parking_plot.setLabel("bottom", "Distance (m)")
        self.parking_plot.addItem(pg.PlotDataItem())
        self.parking_plot.setXRange(min(self.distances), max(self.distances))
        self.parking_plot.setYRange(0, 100)  # Set to standard
        self.parking_plot_curve = self.parking_plot.plot(**self.blue)
        self.parking_smooth_max = et.utils.SmoothMax(self.session_config.update_rate)

    def update_obstruction_text(self) -> None:
        self.obstruction_text_timeout -= 1
        if self.obstruction_text_timeout < 0:
            self.obstruction_text_timeout = 0
            self.obstruction_text_item.hide()

    def show_obstruction_text(self) -> None:
        self.obstruction_text_timeout = 5
        self.obstruction_text_item.show()

    def draw_plot_job(self, *, ref_app_result: RefAppResult) -> None:
        signatures = ref_app_result.extra_result.signature_history
        parking_data = ref_app_result.extra_result.parking_data

        signature_x = [elm[0] for elm in signatures]
        signature_y = [elm[1] for elm in signatures]

        cluster_start = ref_app_result.extra_result.closest_object_dist
        cluster_end = cluster_start + self.cluster_width
        self.sig_plot_curve.setData(x=signature_x, y=signature_y)
        self.sig_plot_cluster_start.setPos(cluster_start)
        self.sig_plot_cluster_end.setPos(cluster_end)
        self.sig_plot_cluster_start.setVisible(False)
        self.sig_plot_cluster_end.setVisible(False)

        sig_max = self.sig_plot_smooth_max.update(max(signature_y))
        self.sig_plot.setYRange(0, sig_max)

        sig_text_x_pos = (
            self.sig_plot_x_range[0] + (self.sig_plot_x_range[1] - self.sig_plot_x_range[0]) * 0.5
        )
        sig_text_y_pos = sig_max * 0.9

        self.parking_car_text_item.setPos(sig_text_x_pos, sig_text_y_pos)
        self.parking_no_car_text_item.setPos(sig_text_x_pos, sig_text_y_pos)

        if ref_app_result.car_detected:
            self.parking_no_car_text_item.hide()
            self.parking_car_text_item.show()
            self.sig_plot_cluster_start.setVisible(True)
            self.sig_plot_cluster_end.setVisible(True)
        else:
            self.parking_car_text_item.hide()
            self.parking_no_car_text_item.show()

        if self.ref_app_config.obstruction_detection:
            obstruction_data = ref_app_result.extra_result.obstruction_data
            self.obstruction_plot_curve.setData(self.obs_distances, obstruction_data)
            point_x, point_y = ref_app_result.extra_result.obstruction_signature
            center_x, center_y = ref_app_result.extra_result.obstruction_center
            rect_x = center_x - self.obs_x_thres
            rect_y = center_y - self.obs_y_thres
            rect_w = 2 * self.obs_x_thres
            rect_h = 2 * self.obs_y_thres

            self.obstruction_center_rect.setRect(rect_x, rect_y, rect_w, rect_h)

            self.obstruction_plot_point.setData(x=[point_x], y=[point_y])
            self.obstruction_plot_center.setData(x=[center_x], y=[center_y])

            if ref_app_result.obstruction_detected:
                self.no_obstruction_text_item.hide()
                self.obstruction_text_item.show()
            else:
                self.obstruction_text_item.hide()
                self.no_obstruction_text_item.show()

        park_max = np.amax(parking_data)
        park_max = self.parking_smooth_max.update(park_max)
        self.parking_plot.setYRange(0, park_max)
        self.parking_plot_curve.setData(self.distances, parking_data)


class ViewPlugin(A121ViewPluginBase):
    TEXT_MSG_MAP = {
        DetailedStatus.OK: "Ready to start.",
        DetailedStatus.CALIBRATION_MISSING: "Run detector calibration.",
        DetailedStatus.CONFIG_MISMATCH: (
            "Current configuration does not match the configuration "
            + "used during detector calibration. Run detector calibration."
        ),
        DetailedStatus.OBSTRUCTION_CALIBRATION_MISSING: "Obstruction calibration missing, run calibration.",
    }

    def __init__(self, app_model: AppModel) -> None:
        super().__init__(app_model=app_model)
        self._log = logging.getLogger(__name__)

        sticky_layout = QVBoxLayout()
        sticky_layout.setContentsMargins(0, 0, 0, 0)
        scrolly_layout = QVBoxLayout()
        scrolly_layout.setContentsMargins(0, 0, 0, 0)

        self.start_button = QPushButton(icons.PLAY(), "Start measurement")
        self.start_button.setShortcut("space")
        self.start_button.setToolTip("Starts the session.\n\nShortcut: Space")
        self.start_button.clicked.connect(self._send_start_request)

        self.stop_button = QPushButton(icons.STOP(), "Stop")
        self.stop_button.setShortcut("space")
        self.stop_button.setToolTip("Stops the session.\n\nShortcut: Space")
        self.stop_button.clicked.connect(self._send_stop_request)

        self.calibrate_ref_app_button = QPushButton(icons.CALIBRATE(), "Calibrate")
        self.calibrate_ref_app_button.setToolTip(
            "Estimate bg noise and (if applicable) calibrates obstruction detection center.\n\n"
        )
        self.calibrate_ref_app_button.clicked.connect(self._on_calibrate_detector)

        self.message_box = QLabel(self.sticky_widget)
        self.message_box.setWordWrap(True)

        button_group = GroupBox.grid("Controls", parent=self.sticky_widget)
        button_group.layout().addWidget(self.start_button, 0, 0)
        button_group.layout().addWidget(self.stop_button, 0, 1)
        button_group.layout().addWidget(self.calibrate_ref_app_button, 1, 0, 1, -1)
        button_group.layout().addWidget(self.message_box, 2, 0, 1, -1)

        sticky_layout.addWidget(button_group)

        self.misc_error_view = MiscErrorView(self.scrolly_widget)
        scrolly_layout.addWidget(self.misc_error_view)

        sensor_selection_group = GroupBox.vertical("Sensor selection", parent=self.scrolly_widget)
        self.sensor_id_pidget = pidgets.SensorIdPidgetFactory(items=[]).create(
            parent=sensor_selection_group
        )
        self.sensor_id_pidget.sig_update.connect(self._on_sensor_id_update)
        sensor_selection_group.layout().addWidget(self.sensor_id_pidget)
        scrolly_layout.addWidget(sensor_selection_group)

        self.parking_config_editor = AttrsConfigEditor(
            config_type=RefAppConfig,
            title="Parking configuration parameters",
            factory_mapping=self._get_parking_pidget_mapping(),
            parent=self.scrolly_widget,
        )
        self.parking_config_editor.sig_update.connect(self._on_parking_config_update)
        scrolly_layout.addWidget(self.parking_config_editor)

        self.obstruction_config_editor = AttrsConfigEditor(
            config_type=RefAppConfig,
            title="Obstruction configuration parameters",
            factory_mapping=self._get_obstruction_pidget_mapping(),
            parent=self.scrolly_widget,
        )
        self.obstruction_config_editor.sig_update.connect(self._on_obstruction_config_update)
        scrolly_layout.addWidget(self.obstruction_config_editor)

        self.sensor_config_editor = AttrsConfigEditor(
            config_type=RefAppConfig,
            title="Sensor configuration",
            factory_mapping=self._get_sensor_config_pidget_mapping(),
            parent=self.scrolly_widget,
        )
        self.sensor_config_editor.sig_update.connect(self._on_config_update)
        scrolly_layout.addWidget(self.sensor_config_editor)

        self.sticky_widget.setLayout(sticky_layout)
        self.scrolly_widget.setLayout(scrolly_layout)

    @classmethod
    def _get_sensor_config_pidget_mapping(cls) -> PidgetGroupFactoryMapping:
        return {
            pidgets.FlatPidgetGroup(): {
                "update_rate": pidgets.FloatPidgetFactory(
                    name_label_text="Update rate:",
                    name_label_tooltip=get_attribute_docstring(RefAppConfig, "update_rate"),
                    suffix=" Hz",
                    decimals=1,
                ),
                "hwaas": pidgets.IntPidgetFactory(
                    name_label_text="HWAAS:",
                    name_label_tooltip=get_attribute_docstring(RefAppConfig, "hwaas"),
                    limits=(1, 511),
                ),
                "profile": pidgets.EnumPidgetFactory(
                    name_label_text="Profile:",
                    name_label_tooltip=get_attribute_docstring(RefAppConfig, "profile"),
                    enum_type=a121.Profile,
                    label_mapping={
                        a121.Profile.PROFILE_1: "1 (shortest)",
                        a121.Profile.PROFILE_2: "2",
                        a121.Profile.PROFILE_3: "3",
                        a121.Profile.PROFILE_4: "4",
                        a121.Profile.PROFILE_5: "5 (longest)",
                    },
                ),
            }
        }

    @classmethod
    def _get_parking_pidget_mapping(cls) -> PidgetGroupFactoryMapping:
        return {
            pidgets.FlatPidgetGroup(): {
                "range_start_m": pidgets.FloatPidgetFactory(
                    name_label_text="Range start:",
                    name_label_tooltip=get_attribute_docstring(RefAppConfig, "range_start_m"),
                    suffix=" m",
                    limits=(0.08, None),
                    decimals=2,
                ),
                "range_end_m": pidgets.FloatPidgetFactory(
                    name_label_text="Range end:",
                    name_label_tooltip=get_attribute_docstring(RefAppConfig, "range_end_m"),
                    suffix=" m",
                    limits=(0.1, None),
                    decimals=2,
                ),
                "queue_length_n": pidgets.IntPidgetFactory(
                    name_label_text="Queue length:",
                    name_label_tooltip=get_attribute_docstring(RefAppConfig, "queue_length_n"),
                    limits=(1, 200),
                ),
                "amplitude_threshold": pidgets.FloatSliderPidgetFactory(
                    name_label_text="Amplitude threshold:",
                    name_label_tooltip=get_attribute_docstring(
                        RefAppConfig, "amplitude_threshold"
                    ),
                    limits=(1.0, 20.0),
                    decimals=1,
                ),
                "weighted_distance_threshold_m": pidgets.FloatPidgetFactory(
                    name_label_text="Weighted distance threshold:",
                    name_label_tooltip=get_attribute_docstring(
                        RefAppConfig, "weighted_distance_threshold_m"
                    ),
                    suffix=" m",
                    limits=(0.0, None),
                    decimals=2,
                ),
                "signature_similarity_threshold": pidgets.FloatSliderPidgetFactory(
                    name_label_text="Signature similarity threshold:",
                    name_label_tooltip=get_attribute_docstring(
                        RefAppConfig, "signature_similarity_threshold"
                    ),
                    suffix=" %",
                    limits=(0, 100.0),
                    decimals=1,
                ),
                "obstruction_detection": pidgets.CheckboxPidgetFactory(
                    name_label_text="Enable obstruction detection",
                    name_label_tooltip=get_attribute_docstring(
                        RefAppConfig, "obstruction_detection"
                    ),
                ),
            }
        }

    @classmethod
    def _get_obstruction_pidget_mapping(cls) -> PidgetGroupFactoryMapping:
        return {
            pidgets.FlatPidgetGroup(): {
                "obstruction_distance_threshold": pidgets.FloatSliderPidgetFactory(
                    name_label_text="Obstruction distance threshold:",
                    name_label_tooltip=get_attribute_docstring(
                        RefAppConfig, "obstruction_distance_threshold"
                    ),
                    limits=(0.001, 1.0),
                    decimals=3,
                    log_scale=True,
                ),
                "obstruction_start_m": pidgets.FloatPidgetFactory(
                    name_label_text="Range start:",
                    name_label_tooltip=get_attribute_docstring(
                        RefAppConfig, "obstruction_start_m"
                    ),
                    suffix=" m",
                    limits=(0.01, None),
                    decimals=2,
                ),
                "obstruction_end_m": pidgets.FloatPidgetFactory(
                    name_label_text="Range end:",
                    name_label_tooltip=get_attribute_docstring(RefAppConfig, "obstruction_end_m"),
                    suffix=" m",
                    limits=(0.01, None),
                    decimals=2,
                ),
            }
        }

    def on_backend_state_update(self, state: Optional[SharedState]) -> None:
        if state is None:
            self.sensor_config_editor.set_data(None)
            self.parking_config_editor.set_data(None)
            self.obstruction_config_editor.set_data(None)
            self.message_box.setText("")
        else:
            self.sensor_id_pidget.set_data(state.sensor_id)
            self.sensor_config_editor.set_data(state.config)
            self.parking_config_editor.set_data(state.config)
            self.obstruction_config_editor.set_data(state.config)

            if state.config is not None and state.context is not None:
                status = RefApp.get_ref_app_status(state.config, state.context)
                self.message_box.setText(self.TEXT_MSG_MAP[status.ref_app_state])
            else:
                self.message_box.setText("")

            results = state.config._collect_validation_results()

            not_handled = self.parking_config_editor.handle_validation_results(results)

            not_handled = self.misc_error_view.handle_validation_results(not_handled)

            assert not_handled == []

    def on_app_model_update(self, app_model: AppModel) -> None:
        main_enable = app_model.plugin_state == PluginState.LOADED_IDLE
        state = app_model.backend_plugin_state

        if state is None:
            ref_app_ready = False
            obs_enable = True
        else:
            obs_enable = state.config.obstruction_detection

            if state.context is not None and state.config is not None:
                status = RefApp.get_ref_app_status(state.config, state.context)
                ref_app_ready = status.ready_to_start
            else:
                ref_app_ready = False

        self.sensor_config_editor.setEnabled(main_enable)
        self.parking_config_editor.setEnabled(main_enable)
        self.obstruction_config_editor.setEnabled(main_enable and obs_enable)

        self.sensor_id_pidget.set_selectable_sensors(app_model.connected_sensors)

        self.calibrate_ref_app_button.setEnabled(
            visual_policies.start_button_enabled(app_model),
        )

        self.start_button.setEnabled(
            visual_policies.start_button_enabled(
                app_model,
                extra_condition=(self.sensor_config_editor.is_ready and ref_app_ready),
            )
        )
        self.stop_button.setEnabled(visual_policies.stop_button_enabled(app_model))

    def _on_config_update(self, config: RefAppConfig) -> None:
        BackendPlugin.update_config.rpc(self.app_model.put_task, config=config)

    def _on_calibrate_detector(self) -> None:
        BackendPlugin.calibrate_detector.rpc(self.app_model.put_task)

    def _on_parking_config_update(self, parking_config: RefAppConfig) -> None:
        config = self.parking_config_editor.get_data()
        assert config is not None
        BackendPlugin.update_config.rpc(self.app_model.put_task, config=config)

    def _on_obstruction_config_update(self, parking_config: RefAppConfig) -> None:
        config = self.obstruction_config_editor.get_data()
        assert config is not None
        BackendPlugin.update_config.rpc(self.app_model.put_task, config=config)

    def _send_defaults_request(self) -> None:
        BackendPlugin.restore_defaults.rpc(self.app_model.put_task)

    def _on_sensor_id_update(self, sensor_id: int) -> None:
        BackendPlugin.update_sensor_id.rpc(self.app_model.put_task, sensor_id=sensor_id)


class PluginSpec(PluginSpecBase):
    def create_backend_plugin(
        self, callback: Callable[[Message], None], key: str
    ) -> BackendPlugin:
        return BackendPlugin(callback=callback, generation=self.generation, key=key)

    def create_view_plugin(self, app_model: AppModel) -> ViewPlugin:
        return ViewPlugin(app_model=app_model)

    def create_plot_plugin(self, app_model: AppModel) -> PlotPlugin:
        return PlotPlugin(app_model=app_model)


PARKING_PLUGIN = PluginSpec(
    generation=PluginGeneration.A121,
    key="parking",
    title="Parking",
    docs_link="https://docs.acconeer.com/en/latest/ref_apps/a121/parking.html",
    description="Detect parked cars.",
    family=PluginFamily.REF_APP,
    presets=[
        PluginPresetBase(
            name="Ground",
            description="Ground mounted",
            preset_id=PluginPresetId.GROUND,
        ),
        PluginPresetBase(
            name="Pole",
            description="Pole mounted",
            preset_id=PluginPresetId.POLE,
        ),
    ],
    default_preset_id=PluginPresetId.GROUND,
)



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

from . import a121, pidgets
from .attrs_config_editor import AttrsConfigEditor, JoinedPartialAttrsConfigEditors
from .collapsible_widget import CollapsibleWidget
from .data_editor import DataEditor
from .goto_resource_tab_button import GotoResourceTabButton
from .group_box import GroupBox
from .json_save_load_buttons import PresentationType, PresenterFunc
from .misc_error_view import MiscErrorView
from .overlay import Overlay
from .tab_pg_widget import TabPGWidget
from .two_sensor_ids_editor import TwoSensorIdsEditor
from .types import PidgetFactoryMapping, PidgetGroupFactoryMapping



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

from . import components, icons, utils
from .components import (
    AttrsConfigEditor,
    DataEditor,
    GroupBox,
    JoinedPartialAttrsConfigEditors,
    MiscErrorView,
    PidgetFactoryMapping,
    PidgetGroupFactoryMapping,
    TwoSensorIdsEditor,
    pidgets,
)



# Copyright (c) Acconeer AB, 2023-2026
# All rights reserved

from __future__ import annotations

import logging
from enum import Enum, auto
from typing import Callable, Mapping, Optional

import attrs
import h5py
import numpy as np

from PySide6.QtGui import QFont
from PySide6.QtWidgets import QPushButton, QVBoxLayout

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool._core.docstrings import get_attribute_docstring
from acconeer.exptool.a121._h5_utils import _create_h5_string_dataset
from acconeer.exptool.a121.algo import APPROX_BASE_STEP_LENGTH_M
from acconeer.exptool.a121.algo._plugins import (
    A121BackendPluginBase,
    A121ViewPluginBase,
)
from acconeer.exptool.a121.algo.breathing import (
    AppState,
    BreathingProcessorConfig,
    get_infant_config,
    get_sitting_config,
)
from acconeer.exptool.a121.algo.presence import (
    DetectorConfig as PresenceDetectorConfig,
)
from acconeer.exptool.a121.algo.presence import (
    ProcessorConfig as PresenceProcessorConfig,
)
from acconeer.exptool.app.new import (
    AppModel,
    AttrsConfigEditor,
    BackendLogger,
    GeneralMessage,
    GroupBox,
    Message,
    MiscErrorView,
    PgPlotPlugin,
    PidgetGroupFactoryMapping,
    PluginFamily,
    PluginGeneration,
    PluginPresetBase,
    PluginSpecBase,
    backend,
    icons,
    is_task,
    pidgets,
    visual_policies,
)
from acconeer.exptool.app.new.ui.components import CollapsibleWidget

from ._ref_app import (
    RefApp,
    RefAppConfig,
    RefAppResult,
    _load_algo_data,
    get_sensor_config,
)


log = logging.getLogger(__name__)


@attrs.mutable(kw_only=True)
class SharedState:
    sensor_id: int = attrs.field(default=1)
    config: RefAppConfig = attrs.field(factory=RefAppConfig)


class PluginPresetId(Enum):
    SITTING = auto()
    INFANT = auto()


@attrs.frozen(kw_only=True)
class SetupMessage(GeneralMessage):
    ref_app_config: RefAppConfig
    sensor_config: a121.SensorConfig
    name: str = attrs.field(default="setup", init=False)
    recipient: backend.RecipientLiteral = attrs.field(default="plot_plugin", init=False)


class BackendPlugin(A121BackendPluginBase[SharedState]):
    PLUGIN_PRESETS: Mapping[int, Callable[[], RefAppConfig]] = {
        PluginPresetId.SITTING.value: lambda: get_sitting_config(),
        PluginPresetId.INFANT.value: lambda: get_infant_config(),
    }

    def __init__(
        self, callback: Callable[[Message], None], generation: PluginGeneration, key: str
    ) -> None:
        super().__init__(callback=callback, generation=generation, key=key)

        self._recorder: Optional[a121.H5Recorder] = None
        self._ref_app_instance: Optional[RefApp] = None
        self._log = BackendLogger.getLogger(__name__)

        self.restore_defaults()

    def _load_from_cache(self, file: h5py.File) -> None:
        self.shared_state.config = RefAppConfig.from_json(file["config"][()])

    @is_task
    def restore_defaults(self) -> None:
        self.shared_state = SharedState()
        self.broadcast()

    @is_task
    def update_sensor_id(self, *, sensor_id: int) -> None:
        self.shared_state.sensor_id = sensor_id
        self.broadcast()

    def _sync_sensor_ids(self) -> None:
        if self.client is not None:
            sensor_ids = self.client.server_info.connected_sensors

            if len(sensor_ids) > 0 and self.shared_state.sensor_id not in sensor_ids:
                self.shared_state.sensor_id = sensor_ids[0]

    @is_task
    def update_config(self, *, config: RefAppConfig) -> None:
        self.shared_state.config = config
        self.broadcast()

    def save_to_cache(self, file: h5py.File) -> None:
        _create_h5_string_dataset(file, "config", self.shared_state.config.to_json())

    @is_task
    def set_preset(self, preset_id: int) -> None:
        preset_config = self.PLUGIN_PRESETS[preset_id]
        self.shared_state.config = preset_config()
        self.broadcast()

    def load_from_record_setup(self, *, record: a121.H5Record) -> None:
        algo_group = record.get_algo_group(self.key)
        _, config = _load_algo_data(algo_group)
        self.shared_state.config = config
        self.shared_state.sensor_id = record.sensor_id

    def _start_session(self, recorder: Optional[a121.H5Recorder]) -> None:
        assert self.client
        self._ref_app_instance = RefApp(
            client=self.client,
            sensor_id=self.shared_state.sensor_id,
            ref_app_config=self.shared_state.config,
        )
        self._ref_app_instance.start(recorder)
        self.callback(
            SetupMessage(
                ref_app_config=self.shared_state.config,
                sensor_config=get_sensor_config(self.shared_state.config),
            )
        )

    def end_session(self) -> None:
        if self._ref_app_instance is None:
            raise RuntimeError
        self._ref_app_instance.stop()

    def get_next(self) -> None:
        assert self.client
        if self._ref_app_instance is None:
            raise RuntimeError
        result = self._ref_app_instance.get_next()

        self.callback(backend.PlotMessage(result=result))


class PlotPlugin(PgPlotPlugin):
    displayed_breathing_rate: Optional[str]

    def __init__(self, app_model: AppModel) -> None:
        super().__init__(app_model=app_model)
        self.displayed_breathing_rate = None
        self._plot_job: Optional[RefAppResult] = None
        self._is_setup = False

    def handle_message(self, message: backend.GeneralMessage) -> None:
        if isinstance(message, backend.PlotMessage):
            self._plot_job = message.result
        elif isinstance(message, SetupMessage):
            self.setup(
                message.ref_app_config,
                message.sensor_config,
            )
            self._is_setup = True
        else:
            log.warn(f"{self.__class__.__name__} got an unsupported command: {message.name!r}.")

    def draw(self) -> None:
        if not self._is_setup or self._plot_job is None:
            return

        try:
            self.draw_plot_job(ref_app_result=self._plot_job)
        finally:
            self._plot_job = None

    def setup(
        self,
        ref_app_config: RefAppConfig,
        sensor_config: a121.SensorConfig,
    ) -> None:
        self.plot_layout.clear()

        self.distances = (
            sensor_config.start_point
            + np.arange(sensor_config.num_points) * sensor_config.step_length
        ) * APPROX_BASE_STEP_LENGTH_M
        self.use_presence_processor = ref_app_config.use_presence_processor

        win = self.plot_layout

        # Define pens and font.
        blue_color = et.utils.color_cycler(0)
        orange_color = et.utils.color_cycler(1)
        brush = et.utils.pg_brush_cycler(0)
        self.blue = dict(
            pen=pg.mkPen(blue_color, width=2),
            symbol="o",
            symbolSize=1,
            symbolBrush=brush,
            symbolPen="k",
        )
        self.orange = dict(
            pen=pg.mkPen(orange_color, width=2),
            symbol="o",
            symbolSize=1,
            symbolBrush=brush,
            symbolPen="k",
        )
        self.blue_transparent_pen = pg.mkPen(f"{blue_color}50", width=2)
        self.orange_transparent_pen = pg.mkPen(f"{orange_color}50", width=2)

        brush_dot = et.utils.pg_brush_cycler(1)
        symbol_dot_kw = dict(symbol="o", symbolSize=10, symbolBrush=brush_dot, symbolPen="k")

        font = QFont()
        font.setPixelSize(16)

        # Presence plot.
        self.presence_plot = win.addPlot(row=0, col=0)
        self.presence_plot.setMenuEnabled(False)
        self.presence_plot.showGrid(x=True, y=True)
        self.presence_plot.addLegend()
        self.presence_plot.setLabel("left", "Presence score")
        self.presence_plot.setLabel("bottom", "Distance (m)")
        self.presence_plot.addItem(pg.PlotDataItem())
        self.presence_plot_curve = []
        self.presence_plot_curve.append(self.presence_plot.plot(**self.blue))
        self.presence_plot_curve.append(self.presence_plot.plot(**self.orange))
        self.presence_plot_curve.append(self.presence_plot.plot(**self.blue))
        self.presence_plot_curve.append(self.presence_plot.plot(**self.orange))

        self.presence_plot_legend = pg.LegendItem(offset=(0.0, 0.5))
        self.presence_plot_legend.setParentItem(self.presence_plot)
        self.presence_plot_legend.addItem(self.presence_plot_curve[2], "Slow motion")
        self.presence_plot_legend.addItem(self.presence_plot_curve[3], "Fast motion")
        self.presence_plot_legend.show()

        self.presence_smoot_max = et.utils.SmoothMax()

        self.presence_text_item = pg.TextItem(
            fill=pg.mkColor(0xFF, 0x7F, 0x0E, 200),
            anchor=(0.5, 0),
            color=pg.mkColor(0xFF, 0xFF, 0xFF, 200),
        )
        self.presence_text_item.setFont(font)
        self.presence_text_item.show()
        self.presence_plot.addItem(self.presence_text_item)

        # Time series plot.
        self.time_series_plot = win.addPlot(row=1, col=0)
        self.time_series_plot.setMenuEnabled(False)
        self.time_series_plot.showGrid(x=True, y=True)
        self.time_series_plot.addLegend()
        self.time_series_plot.setLabel("left", "Displacement")
        self.time_series_plot.setLabel("bottom", "Time (s)")
        self.time_series_plot.addItem(pg.PlotDataItem())
        self.time_series_curve = self.time_series_plot.plot(**self.blue)

        self.time_series_text_item = pg.TextItem(
            fill=pg.mkColor(0xFF, 0x7F, 0x0E, 200),
            anchor=(0.5, 0),
            color=pg.mkColor(0xFF, 0xFF, 0xFF, 200),
        )
        self.time_series_text_item.setFont(font)
        self.time_series_text_item.show()
        self.time_series_plot.addItem(self.time_series_text_item)

        # Breathing psd plot.
        self.breathing_psd_plot = win.addPlot(row=2, col=0)
        self.breathing_psd_plot.setMenuEnabled(False)
        self.breathing_psd_plot.showGrid(x=True, y=True)
        self.breathing_psd_plot.addLegend()
        self.breathing_psd_plot.setLabel("left", "PSD")
        self.breathing_psd_plot.setLabel("bottom", "Breathing rate (Hz)")
        self.breathing_psd_plot.addItem(pg.PlotDataItem())
        self.breathing_psd_curve = self.breathing_psd_plot.plot(**self.blue)

        self.psd_smoothing = et.utils.SmoothMax()

        # Breathing rate plot.
        self.breathing_rate_plot = win.addPlot(row=3, col=0)
        self.breathing_rate_plot.setMenuEnabled(False)
        self.breathing_rate_plot.showGrid(x=True, y=True)
        self.breathing_rate_plot.addLegend()
        self.breathing_rate_plot.setLabel("left", "Breaths per minute")
        self.breathing_rate_plot.setLabel("bottom", "Time (s)")
        self.breathing_rate_plot.addItem(pg.PlotDataItem())
        self.breathing_rate_curves = []
        self.breathing_rate_curves.append(self.breathing_rate_plot.plot(**self.blue))
        self.breathing_rate_curves.append(
            self.breathing_rate_plot.plot(**dict(pen=None, **symbol_dot_kw))
        )
        self.smooth_breathing_rate = et.utils.SmoothLimits()

        self.breathing_rate_plot_legend = pg.LegendItem(offset=(0.0, 0.5))
        self.breathing_rate_plot_legend.setParentItem(self.breathing_rate_plot)
        self.breathing_rate_plot_legend.addItem(self.breathing_rate_curves[0], "Breathing rate")
        self.breathing_rate_plot_legend.addItem(
            self.breathing_rate_curves[1], "Breathing rate(embedded output)"
        )
        self.breathing_rate_plot_legend.hide()

        self.breathing_rate_text_item = pg.TextItem(
            fill=pg.mkColor(0xFF, 0x7F, 0x0E, 200),
            anchor=(0.5, 0),
            color=pg.mkColor(0xFF, 0xFF, 0xFF, 200),
        )
        self.breathing_rate_text_item.setFont(font)
        self.breathing_rate_text_item.hide()
        self.breathing_rate_plot.addItem(self.breathing_rate_text_item)

    def draw_plot_job(self, *, ref_app_result: RefAppResult) -> None:
        app_state = ref_app_result.app_state

        max_ampl = max(
            np.max(ref_app_result.presence_result.inter),
            np.max(ref_app_result.presence_result.intra),
        )
        lim = self.presence_smoot_max.update(max_ampl)
        self.presence_plot.setYRange(0, lim)

        if ref_app_result.distances_being_analyzed is None:
            self.presence_plot_curve[0].setData(
                self.distances, ref_app_result.presence_result.inter, **self.blue
            )
            self.presence_plot_curve[1].setData(
                self.distances, ref_app_result.presence_result.intra, **self.orange
            )
            self.presence_plot_curve[2].setData([], [])
            self.presence_plot_curve[3].setData([], [])
        else:
            start = ref_app_result.distances_being_analyzed[0]
            end = ref_app_result.distances_being_analyzed[1]
            s = slice(start, end)
            distance_slice = self.distances[s]
            self.presence_plot_curve[0].setData(
                self.distances,
                ref_app_result.presence_result.inter,
                pen=self.blue_transparent_pen,
            )
            self.presence_plot_curve[1].setData(
                self.distances,
                ref_app_result.presence_result.intra,
                pen=self.orange_transparent_pen,
            )
            self.presence_plot_curve[2].setData(
                distance_slice, ref_app_result.presence_result.inter[s]
            )
            self.presence_plot_curve[3].setData(
                distance_slice, ref_app_result.presence_result.intra[s]
            )

        if ref_app_result.breathing_result is not None:
            breathing_result = ref_app_result.breathing_result.extra_result
            breathing_motion = breathing_result.breathing_motion
            psd = breathing_result.psd
            frequencies = breathing_result.frequencies
            time_vector = breathing_result.time_vector
            all_breathing_rate_history = breathing_result.all_breathing_rate_history
            breathing_rate_history = breathing_result.breathing_rate_history

            self.time_series_curve.setData(
                time_vector[-breathing_motion.shape[0] :], breathing_motion
            )
            y = np.max(np.abs(breathing_motion)) * 1.05
            self.time_series_plot.setYRange(-y, y)
            self.time_series_plot.setXRange(
                time_vector[-breathing_motion.shape[0]], max(time_vector)
            )

            if not np.all(np.isnan(all_breathing_rate_history)):
                ylim = self.psd_smoothing.update(psd)
                self.breathing_psd_curve.setData(frequencies, psd)
                self.breathing_psd_plot.setYRange(0, ylim)
                self.breathing_psd_plot.setXRange(0, 2)

                self.breathing_rate_curves[0].setData(time_vector, all_breathing_rate_history)
                lims = self.smooth_breathing_rate.update(all_breathing_rate_history)
                self.breathing_rate_plot.setYRange(lims[0] - 3, lims[1] + 3)

                self.breathing_rate_plot_legend.show()

            if not np.all(np.isnan(breathing_rate_history)):
                self.breathing_rate_curves[1].setData(time_vector, breathing_rate_history)

            if not np.isnan(breathing_rate_history[-1]):
                self.displayed_breathing_rate = "{:.1f}".format(breathing_rate_history[-1])
                self.breathing_rate_text_item.show()

        else:
            self.time_series_plot.setYRange(0, 1)
            self.time_series_plot.setXRange(0, 1)
            self.time_series_curve.setData([], [])
            self.breathing_psd_curve.setData([], [])
            self.breathing_rate_curves[0].setData([], [])
            self.breathing_rate_curves[1].setData([], [])
            self.displayed_breathing_rate = None
            self.breathing_rate_text_item.hide()

        # Set text in text boxes according to app state.

        # Presence text
        if app_state == AppState.NO_PRESENCE_DETECTED:
            presence_text = "No presence detected"
        elif app_state == AppState.DETERMINE_DISTANCE_ESTIMATE:
            presence_text = "Determining distance with presence"
        elif app_state == AppState.ESTIMATE_BREATHING_RATE:
            start_m = "{:.2f}".format(distance_slice[0])
            end_m = "{:.2f}".format(distance_slice[-1])
            if self.use_presence_processor:
                presence_text = (
                    "Presence detected in the range " + start_m + " - " + end_m + " (m)"
                )
            else:
                presence_text = "Presence distance detection disabled"
        elif app_state == AppState.INTRA_PRESENCE_DETECTED:
            presence_text = "Large motion detected"
        else:
            presence_text = ""

        text_y_pos = self.presence_plot.getAxis("left").range[1] * 0.95
        text_x_pos = (
            self.presence_plot.getAxis("bottom").range[1]
            + self.presence_plot.getAxis("bottom").range[0]
        ) / 2.0
        self.presence_text_item.setPos(text_x_pos, text_y_pos)
        self.presence_text_item.setHtml(presence_text)

        # Breathing text
        if app_state == AppState.ESTIMATE_BREATHING_RATE:
            if (
                ref_app_result.breathing_result is not None
                and ref_app_result.breathing_result.breathing_rate is None
            ):
                time_series_text = "Initializing breathing detection"
            elif self.displayed_breathing_rate is not None:
                time_series_text = "Breathing rate: " + self.displayed_breathing_rate + " bpm"
            else:
                time_series_text = ""
        else:
            time_series_text = "Waiting for distance"

        text_y_pos = self.time_series_plot.getAxis("left").range[1] * 0.95
        text_x_pos = (
            self.time_series_plot.getAxis("bottom").range[1]
            + self.time_series_plot.getAxis("bottom").range[0]
        ) / 2.0
        self.time_series_text_item.setPos(text_x_pos, text_y_pos)
        self.time_series_text_item.setHtml(time_series_text)

        if self.displayed_breathing_rate is not None:
            text_y_pos = self.breathing_rate_plot.getAxis("left").range[1] * 0.95
            text_x_pos = time_vector[0]

            self.breathing_rate_text_item.setPos(text_x_pos, text_y_pos)
            self.breathing_rate_text_item.setHtml(self.displayed_breathing_rate + " bpm")


class ViewPlugin(A121ViewPluginBase):
    def __init__(self, app_model: AppModel) -> None:
        super().__init__(app_model=app_model)
        self._log = logging.getLogger(__name__)

        sticky_layout = QVBoxLayout()
        sticky_layout.setContentsMargins(0, 0, 0, 0)
        scrolly_layout = QVBoxLayout()
        scrolly_layout.setContentsMargins(0, 0, 0, 0)

        self.start_button = QPushButton(icons.PLAY(), "Start measurement")
        self.start_button.setShortcut("space")
        self.start_button.setToolTip("Starts the session.\n\nShortcut: Space")
        self.start_button.clicked.connect(self._send_start_request)

        self.stop_button = QPushButton(icons.STOP(), "Stop")
        self.stop_button.setShortcut("space")
        self.stop_button.setToolTip("Stops the session.\n\nShortcut: Space")
        self.stop_button.clicked.connect(self._send_stop_request)

        button_group = GroupBox.grid("Controls", parent=self.sticky_widget)
        button_group.layout().addWidget(self.start_button, 0, 0)
        button_group.layout().addWidget(self.stop_button, 0, 1)

        sticky_layout.addWidget(button_group)

        self.misc_error_view = MiscErrorView(self.scrolly_widget)
        scrolly_layout.addWidget(self.misc_error_view)

        sensor_selection_group = GroupBox.vertical("Sensor selection", parent=self.scrolly_widget)
        self.sensor_id_pidget = pidgets.SensorIdPidgetFactory(items=[]).create(
            parent=sensor_selection_group
        )
        self.sensor_id_pidget.sig_update.connect(self._on_sensor_id_update)
        sensor_selection_group.layout().addWidget(self.sensor_id_pidget)
        scrolly_layout.addWidget(sensor_selection_group)

        self.config_editor = AttrsConfigEditor(
            config_type=RefAppConfig,
            title="Ref App parameters",
            factory_mapping=self._get_ref_app_pidget_mapping(),
            parent=self.scrolly_widget,
        )
        self.config_editor.sig_update.connect(self._on_config_update)
        scrolly_layout.addWidget(self.config_editor)

        self.breathing_config_editor = AttrsConfigEditor(
            config_type=BreathingProcessorConfig,
            title="Breathing configuration parameters",
            factory_mapping=self._get_breathing_pidget_mapping(),
            parent=self.scrolly_widget,
        )
        self.breathing_config_editor.sig_update.connect(self._on_breathing_config_update)
        scrolly_layout.addWidget(self.breathing_config_editor)

        self.sensor_config_editor = AttrsConfigEditor(
            config_type=RefAppConfig,
            title="Sensor configuration",
            factory_mapping=self._get_sensor_config_pidget_mapping(),
            parent=self.scrolly_widget,
        )
        self.sensor_config_editor.sig_update.connect(self._on_config_update)
        scrolly_layout.addWidget(self.sensor_config_editor)

        self.presence_config_editor = AttrsConfigEditor(
            config_type=PresenceProcessorConfig,
            title="Presence configuration parameters",
            factory_mapping=self._get_presence_config_pidget_mapping(),
            parent=self.scrolly_widget,
        )
        self.presence_config_editor.sig_update.connect(self._on_presence_config_update)

        self.collapsible_widget = CollapsibleWidget(
            "Presence configuration parameters", self.presence_config_editor, self.scrolly_widget
        )
        scrolly_layout.addWidget(self.collapsible_widget)

        self.sticky_widget.setLayout(sticky_layout)
        self.scrolly_widget.setLayout(scrolly_layout)

    @classmethod
    def _get_ref_app_pidget_mapping(cls) -> PidgetGroupFactoryMapping:
        return {
            pidgets.FlatPidgetGroup(): {
                "start_m": pidgets.FloatPidgetFactory(
                    name_label_text="Start:",
                    suffix=" m",
                    decimals=1,
                    limits=(0, None),
                    name_label_tooltip=get_attribute_docstring(RefAppConfig, "start_m"),
                ),
                "end_m": pidgets.FloatPidgetFactory(
                    name_label_text="End:",
                    suffix=" m",
                    decimals=1,
                    limits=(0, None),
                    name_label_tooltip=get_attribute_docstring(RefAppConfig, "end_m"),
                ),
                "num_distances_to_analyze": pidgets.IntPidgetFactory(
                    name_label_text="Number of distances to analyze:",
                    limits=(1, 64),
                    name_label_tooltip=get_attribute_docstring(
                        RefAppConfig, "num_distances_to_analyze"
                    ),
                ),
                "distance_determination_duration": pidgets.FloatPidgetFactory(
                    name_label_text="Duration to determine distance:",
                    suffix=" s",
                    decimals=1,
                    limits=(1, None),
                    name_label_tooltip=get_attribute_docstring(
                        RefAppConfig, "distance_determination_duration"
                    ),
                ),
                "use_presence_processor": pidgets.CheckboxPidgetFactory(
                    name_label_text="Use presence processor to determine distance",
                    name_label_tooltip=get_attribute_docstring(
                        RefAppConfig, "use_presence_processor"
                    ),
                ),
            }
        }

    @classmethod
    def _get_sensor_config_pidget_mapping(cls) -> PidgetGroupFactoryMapping:
        return {
            pidgets.FlatPidgetGroup(): {
                "frame_rate": pidgets.FloatPidgetFactory(
                    name_label_text="Frame rate:",
                    suffix=" Hz",
                    decimals=1,
                    name_label_tooltip=get_attribute_docstring(RefAppConfig, "frame_rate"),
                ),
                "sweeps_per_frame": pidgets.IntPidgetFactory(
                    name_label_text="Sweeps per frame:",
                    limits=(1, 64),
                    name_label_tooltip=get_attribute_docstring(RefAppConfig, "sweeps_per_frame"),
                ),
                "hwaas": pidgets.IntPidgetFactory(
                    name_label_text="HWAAS:",
                    limits=(1, 511),
                    name_label_tooltip=get_attribute_docstring(RefAppConfig, "hwaas"),
                ),
                "profile": pidgets.EnumPidgetFactory(
                    name_label_text="Profile:",
                    enum_type=a121.Profile,
                    label_mapping={
                        a121.Profile.PROFILE_1: "1 (shortest)",
                        a121.Profile.PROFILE_2: "2",
                        a121.Profile.PROFILE_3: "3",
                        a121.Profile.PROFILE_4: "4",
                        a121.Profile.PROFILE_5: "5 (longest)",
                    },
                    name_label_tooltip=get_attribute_docstring(RefAppConfig, "profile"),
                ),
            }
        }

    @classmethod
    def _get_breathing_pidget_mapping(cls) -> PidgetGroupFactoryMapping:
        return {
            pidgets.FlatPidgetGroup(): {
                "lowest_breathing_rate": pidgets.FloatPidgetFactory(
                    name_label_text="Lowest anticipated breathing rate:",
                    suffix=" bpm",
                    limits=(2, None),
                    decimals=1,
                    name_label_tooltip=get_attribute_docstring(
                        BreathingProcessorConfig, "lowest_breathing_rate"
                    ),
                ),
                "highest_breathing_rate": pidgets.FloatPidgetFactory(
                    name_label_text="Highest anticipated breathing rate:",
                    suffix=" bpm",
                    limits=(2, None),
                    decimals=1,
                    name_label_tooltip=get_attribute_docstring(
                        BreathingProcessorConfig, "highest_breathing_rate"
                    ),
                ),
                "time_series_length_s": pidgets.FloatPidgetFactory(
                    name_label_text="Time series length:",
                    suffix=" s",
                    decimals=1,
                    name_label_tooltip=get_attribute_docstring(
                        BreathingProcessorConfig, "time_series_length_s"
                    ),
                ),
            }
        }

    @classmethod
    def _get_presence_config_pidget_mapping(cls) -> PidgetGroupFactoryMapping:
        return {
            pidgets.FlatPidgetGroup(): {
                "intra_detection_threshold": pidgets.FloatSliderPidgetFactory(
                    name_label_text="Intra detection threshold:",
                    name_label_tooltip=get_attribute_docstring(
                        PresenceDetectorConfig,  # DetectorConfig is not used, but has all the descriptions
                        "intra_detection_threshold",
                    ),
                    decimals=2,
                    limits=(0, 15),
                ),
                "intra_frame_time_const": pidgets.FloatSliderPidgetFactory(
                    name_label_text="Intra time constant:",
                    name_label_tooltip=get_attribute_docstring(
                        PresenceDetectorConfig,  # DetectorConfig is not used, but has all the descriptions
                        "intra_frame_time_const",
                    ),
                    suffix=" s",
                    decimals=2,
                    limits=(0, 1),
                ),
                "intra_output_time_const": pidgets.FloatSliderPidgetFactory(
                    name_label_text="Intra output time constant:",
                    name_label_tooltip=get_attribute_docstring(
                        PresenceDetectorConfig,  # DetectorConfig is not used, but has all the descriptions
                        "intra_output_time_const",
                    ),
                    suffix=" s",
                    decimals=2,
                    limits=(0.01, 20),
                    log_scale=True,
                ),
                "inter_detection_threshold": pidgets.FloatSliderPidgetFactory(
                    name_label_text="Inter detection threshold:",
                    name_label_tooltip=get_attribute_docstring(
                        PresenceDetectorConfig,  # DetectorConfig is not used, but has all the descriptions
                        "inter_detection_threshold",
                    ),
                    decimals=2,
                    limits=(0, 5),
                ),
                "inter_frame_fast_cutoff": pidgets.FloatSliderPidgetFactory(
                    name_label_text="Inter fast cutoff freq.:",
                    name_label_tooltip=get_attribute_docstring(
                        PresenceDetectorConfig,  # DetectorConfig is not used, but has all the descriptions
                        "inter_frame_fast_cutoff",
                    ),
                    suffix=" Hz",
                    decimals=2,
                    limits=(1, 50),
                    log_scale=True,
                ),
                "inter_frame_slow_cutoff": pidgets.FloatSliderPidgetFactory(
                    name_label_text="Inter slow cutoff freq.:",
                    name_label_tooltip=get_attribute_docstring(
                        PresenceDetectorConfig,  # DetectorConfig is not used, but has all the descriptions
                        "inter_frame_slow_cutoff",
                    ),
                    suffix=" Hz",
                    decimals=2,
                    limits=(0.01, 1),
                    log_scale=True,
                ),
                "inter_frame_deviation_time_const": pidgets.FloatSliderPidgetFactory(
                    name_label_text="Inter time constant:",
                    name_label_tooltip=get_attribute_docstring(
                        PresenceDetectorConfig,  # DetectorConfig is not used, but has all the descriptions
                        "inter_frame_deviation_time_const",
                    ),
                    suffix=" s",
                    decimals=2,
                    limits=(0.01, 20),
                    log_scale=True,
                ),
                "inter_output_time_const": pidgets.FloatSliderPidgetFactory(
                    name_label_text="Inter output time constant:",
                    name_label_tooltip=get_attribute_docstring(
                        PresenceDetectorConfig,  # DetectorConfig is not used, but has all the descriptions
                        "inter_output_time_const",
                    ),
                    suffix=" s",
                    decimals=2,
                    limits=(0.01, 20),
                    log_scale=True,
                ),
            }
        }

    def on_backend_state_update(self, state: Optional[SharedState]) -> None:
        if state is None:
            self.config_editor.set_data(None)
            self.sensor_config_editor.set_data(None)
            self.breathing_config_editor.set_data(None)
            self.presence_config_editor.set_data(None)
        else:
            self.sensor_id_pidget.set_data(state.sensor_id)
            self.config_editor.set_data(state.config)
            self.sensor_config_editor.set_data(state.config)
            self.breathing_config_editor.set_data(state.config.breathing_config)
            self.presence_config_editor.set_data(state.config.presence_config)

            results = state.config._collect_validation_results()

            not_handled = self.config_editor.handle_validation_results(results)

            not_handled = self.misc_error_view.handle_validation_results(not_handled)

            assert not_handled == []

    def on_app_model_update(self, app_model: AppModel) -> None:
        self.sensor_id_pidget.set_selectable_sensors(app_model.connected_sensors)

        visual_policies.apply_enabled_policy(
            visual_policies.config_editor_enabled,
            app_model,
            widgets=[
                self.sensor_id_pidget,
                self.config_editor,
                self.sensor_config_editor,
                self.breathing_config_editor,
                self.presence_config_editor,
            ],
        )
        self.start_button.setEnabled(
            visual_policies.start_button_enabled(
                app_model, extra_condition=self.config_editor.is_ready
            )
        )
        self.stop_button.setEnabled(visual_policies.stop_button_enabled(app_model))

    def _on_config_update(self, config: RefAppConfig) -> None:
        BackendPlugin.update_config.rpc(self.app_model.put_task, config=config)

    def _on_breathing_config_update(self, breathing_config: BreathingProcessorConfig) -> None:
        config = self.config_editor.get_data()
        assert config is not None
        config.breathing_config = breathing_config
        BackendPlugin.update_config.rpc(self.app_model.put_task, config=config)

    def _on_presence_config_update(self, presence_config: PresenceProcessorConfig) -> None:
        config = self.config_editor.get_data()
        assert config is not None
        config.presence_config = presence_config
        BackendPlugin.update_config.rpc(self.app_model.put_task, config=config)

    def _send_defaults_request(self) -> None:
        BackendPlugin.restore_defaults.rpc(self.app_model.put_task)

    def _on_sensor_id_update(self, sensor_id: int) -> None:
        BackendPlugin.update_sensor_id.rpc(self.app_model.put_task, sensor_id=sensor_id)


class PluginSpec(PluginSpecBase):
    def create_backend_plugin(
        self, callback: Callable[[Message], None], key: str
    ) -> BackendPlugin:
        return BackendPlugin(callback=callback, generation=self.generation, key=key)

    def create_view_plugin(self, app_model: AppModel) -> ViewPlugin:
        return ViewPlugin(app_model=app_model)

    def create_plot_plugin(self, app_model: AppModel) -> PlotPlugin:
        return PlotPlugin(app_model=app_model)


BREATHING_PLUGIN = PluginSpec(
    generation=PluginGeneration.A121,
    key="breathing",
    title="Breathing",
    docs_link="https://docs.acconeer.com/en/latest/ref_apps/a121/breathing.html",
    description="Detect breathing rate.",
    family=PluginFamily.REF_APP,
    presets=[
        PluginPresetBase(
            name="Sitting",
            description="Sitting",
            preset_id=PluginPresetId.SITTING,
        ),
        PluginPresetBase(
            name="Infant",
            description="Infant",
            preset_id=PluginPresetId.INFANT,
        ),
    ],
    default_preset_id=PluginPresetId.SITTING,
)



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

import json
from typing import Any, Optional

import attrs
import packaging.version

from acconeer.exptool._core.class_creation.formatting import pretty_dict_line_strs
from acconeer.exptool.a121._core import utils


@attrs.frozen(kw_only=True)
class SensorInfo:
    """Holds information about a single sensor slot on a host."""

    connected: bool
    serial: Optional[str] = None

    def to_dict(self) -> dict[str, Any]:
        return attrs.asdict(self)

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> SensorInfo:
        return cls(**d)

    def to_json(self) -> str:
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> SensorInfo:
        return cls.from_dict(json.loads(json_str))

    def _pretty_str_lines(self, index: Optional[int] = None) -> list[str]:
        index_str = "" if index is None else f" @ slot {index}"
        return [
            f"{type(self).__name__}{index_str}:",
            *pretty_dict_line_strs(self.to_dict()),
        ]

    def __str__(self) -> str:
        return "\n".join(self._pretty_str_lines())


@attrs.frozen(kw_only=True)
class ServerInfo:
    rss_version: str = attrs.field()
    sensor_count: int = attrs.field()
    ticks_per_second: int = attrs.field()
    sensor_infos: dict[int, SensorInfo] = attrs.field()
    hardware_name: Optional[str] = attrs.field(default=None)
    max_baudrate: Optional[int] = attrs.field(default=None)

    @property
    def parsed_rss_version(self) -> packaging.version.Version:
        return utils.parse_rss_version(self.rss_version)

    @property
    def connected_sensors(self) -> list[int]:
        return [
            sensor_id
            for sensor_id, sensor_info in self.sensor_infos.items()
            if sensor_info.connected
        ]

    def to_dict(self) -> dict[str, Any]:
        return attrs.asdict(self)

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> ServerInfo:
        d = d.copy()
        d["sensor_infos"] = {
            int(sensor_id): SensorInfo.from_dict(sensor_info_dict)
            for sensor_id, sensor_info_dict in d["sensor_infos"].items()
        }
        return cls(**d)

    def to_json(self) -> str:
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> ServerInfo:
        return cls.from_dict(json.loads(json_str))

    def __str__(self) -> str:
        d = self.to_dict()
        del d["sensor_infos"]

        sensor_infos_strs = [
            line
            for i, sensor_info in self.sensor_infos.items()
            for line in utils.indent_strs(sensor_info._pretty_str_lines(index=i), 2)
        ]

        return "\n".join(
            [
                f"{type(self).__name__}:",
                *pretty_dict_line_strs(d),
                "  sensor_infos:",
                *sensor_infos_strs,
            ]
        )



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from __future__ import annotations

import re

import typing_extensions as te

from PySide6.QtGui import QColor, QSyntaxHighlighter, QTextCharFormat, QTextDocument


class _Rgb(te.TypedDict, total=True):
    r: int
    g: int
    b: int


class SyntaxHighlighter(QSyntaxHighlighter):
    GOLDEN_ROD: _Rgb = dict(r=218, g=165, b=32)
    PURPLE: _Rgb = dict(r=160, g=32, b=240)
    FOREST_GREEN: _Rgb = dict(r=34, g=139, b=34)
    DODGER_BLUE: _Rgb = dict(r=30, g=144, b=255)
    FIREBRICK: _Rgb = dict(r=178, g=34, b=34)
    SIENNA: _Rgb = dict(r=160, g=82, b=45)
    GRAY: _Rgb = dict(r=100, g=100, b=100)

    def __init__(self, parent: QTextDocument, highlights: dict[str, QTextCharFormat]) -> None:
        super().__init__(parent)
        self._highlights = highlights

    def highlightBlock(self, text: str) -> None:
        for pattern, fmt in self._highlights.items():
            for match in re.finditer(pattern, text):
                # The first capturing group in the regex will be styled
                try:
                    match_length = match.end(1) - match.start(1)
                    self.setFormat(match.start(1), match_length, fmt)
                except IndexError:
                    pass

    @staticmethod
    def _create_char_fmt(
        *,
        r: int = 0,
        g: int = 0,
        b: int = 0,
        italic: bool = False,
        weight: int = 400,
    ) -> QTextCharFormat:
        fmt = QTextCharFormat()
        fmt.setForeground(QColor(r, g, b))
        fmt.setFontItalic(italic)
        fmt.setFontWeight(weight)
        return fmt

    @classmethod
    def json(cls, parent: QTextDocument) -> SyntaxHighlighter:
        return cls(
            parent,
            {
                r'("[^"]*")': cls._create_char_fmt(**cls.GOLDEN_ROD),  # string literals
                r"(\d+(?:\.\d+)?),": cls._create_char_fmt(**cls.DODGER_BLUE),  # number literals
            },
        )

    @classmethod
    def rss_c(cls, parent: QTextDocument) -> SyntaxHighlighter:
        snake_case = "[a-z]+(?:_[a-z]+)*"
        type_regex = f"{snake_case}_t "
        acc_type_regex = f"acc_{type_regex}"
        return cls(
            parent,
            {
                # anything snake_case
                rf"({snake_case})": cls._create_char_fmt(**cls.SIENNA),
                # keywords
                r"(static)": cls._create_char_fmt(**cls.PURPLE),
                # types
                rf"(void|{type_regex}|{acc_type_regex})": cls._create_char_fmt(**cls.FOREST_GREEN),
                # function call & -definition
                rf"(set_config|acc_{snake_case}_set)\(": cls._create_char_fmt(b=255, weight=500),
                # int literals
                r"(\d+U?)": cls._create_char_fmt(**cls.DODGER_BLUE),
                # floating literals
                r"(\d+\.\d+f?)": cls._create_char_fmt(**cls.DODGER_BLUE),
                # bool literals
                "(true|false)": cls._create_char_fmt(r=0, g=0, b=0),
                # acconeer enum members/constants
                r"(ACC[_A-Z0-9]*)": cls._create_char_fmt(**cls.GOLDEN_ROD),
                # single line comments
                r"(//.*)": cls._create_char_fmt(**cls.FIREBRICK, italic=True),
                # parameter ignores
                rf"(\(void\){snake_case};)": cls._create_char_fmt(**cls.GRAY, italic=True),
            },
        )



# Copyright (c) Acconeer AB, 2023-2024
# All rights reserved

from ._detector import DetailedStatus, Detector, DetectorConfig, DetectorContext, DetectorResult
from ._processors import (
    Processor,
    ProcessorConfig,
    ProcessorContext,
    ProcessorResult,
    Target,
)



# Copyright (c) Acconeer AB, 2025
# All rights reserved

from ._ex_app import (
    CargoPresenceConfig,
    ContainerSize,
    ExApp,
    ExAppConfig,
    ExAppContext,
    ExAppResult,
    UtilizationLevelConfig,
)



# Copyright (c) Acconeer AB, 2022-2025
# All rights reserved

from __future__ import annotations

import logging
from typing import List, Optional

import bs4
import requests


DEV_LICENSE_URL = "https://developer.acconeer.com/software-license-agreement/"

DEV_LICENSE_DEFAULT_HEADER = "SOFTWARE LICENSE AGREEMENT"
DEV_LICENSE_DEFAULT_SUBHEADER = "LIMITED LICENSE AGREEMENT FOR Acconeer MATERIALS"
DEV_LICENSE_DEFAULT_CONFIRMATION_MSG = (
    "Please read carefully the terms and conditions at "
    "https://developer.acconeer.com/software-license-agreement/"
)

log = logging.getLogger(__name__)


class DevLicense:
    """Acconeer developer license abstraction"""

    def __init__(self, license_url: str = DEV_LICENSE_URL) -> None:
        self.license_url = license_url
        self.html: Optional[bs4.BeautifulSoup] = None

    def load(self) -> None:
        """Loads the license HTML document"""
        try:
            self.html = bs4.BeautifulSoup(requests.get(self.license_url).text, "html.parser")
        except Exception as e:
            log.warn(str(e))
            self.html = None

    def get_header(self) -> str:
        """Get the HTML header as a string

        :returns: HTML header.

        """
        if self.html is None:
            return DEV_LICENSE_DEFAULT_HEADER

        last_breadcrumb = self.html.find("span", {"class": "breadcrumb_last"})
        if not isinstance(last_breadcrumb, bs4.Tag):
            return DEV_LICENSE_DEFAULT_HEADER

        header_text = last_breadcrumb.string
        if header_text is None:
            return DEV_LICENSE_DEFAULT_HEADER

        return header_text

    def get_header_element(self) -> str:
        """Get the HTML header element as a string

        :returns: HTML header element.

        """
        return "<h1>" + self.get_header() + "</h1>"

    def get_subheader(self) -> str:
        """Get the HTML subheader as a string

        :returns: HTML subheader.

        """
        subheader = DEV_LICENSE_DEFAULT_SUBHEADER

        if self.html is not None:
            subheader_element = self.html.find("h3")
            if isinstance(subheader_element, bs4.Tag):
                h3_contents = subheader_element.string
                if h3_contents is not None:
                    subheader = h3_contents

        return subheader

    def get_subheader_element(self) -> str:
        """Get the HTML subheader element as a string

        :returns: HTML subheader element.

        """
        subheader_elem = "<h3>" + DEV_LICENSE_DEFAULT_SUBHEADER + "</h3>"

        if self.html is not None:
            subheader_elem = str(self.html.find("h3"))

        return subheader_elem

    def get_content(self) -> List[str]:
        """Get the HTML paragraph content as a list of strings

        :returns: HTML paragraphs.

        """
        paragraphs = []

        if self.html is not None:
            for p in self.html.select("p"):
                if p is not None and p.string is not None:
                    # The "\xa0" is a non-breaking space (&nbsp;) that is not
                    # escaped by BeautifulSoup and hence filtered out here.
                    paragraphs.append(p.string.replace("\xa0", " "))
        else:
            paragraphs.append(DEV_LICENSE_DEFAULT_CONFIRMATION_MSG)

        return paragraphs

    def get_content_elements(self) -> List[str]:
        """Get the HTML paragraph elements as a list of strings

        :returns: HTML paragraphs elements.

        """
        paragraphs = []

        if self.html is not None:
            paragraphs = [str(p) for p in self.html.select("p") if "class=" not in str(p)]
        else:
            paragraphs.append("<p></p>")
            paragraphs.append(DEV_LICENSE_DEFAULT_CONFIRMATION_MSG)

        return paragraphs



# Copyright (c) Acconeer AB, 2023-2025
# All rights reserved

from acconeer.exptool import a121
from acconeer.exptool.a121.algo.distance import PeakSortingMethod, ReflectorShape, ThresholdMethod
from acconeer.exptool.a121.algo.tank_level._ref_app import RefAppConfig


def get_small_config() -> RefAppConfig:
    return RefAppConfig(
        median_filter_length=3,
        num_medians_to_average=2,
        start_m=0.030,
        end_m=0.5,
        max_step_length=12,
        max_profile=a121.Profile.PROFILE_3,
        num_frames_in_recorded_threshold=50,
        peaksorting_method=PeakSortingMethod.CLOSEST,
        reflector_shape=ReflectorShape.PLANAR,
        threshold_method=ThresholdMethod.CFAR,
        close_range_leakage_cancellation=False,
        fixed_threshold_value=100.0,
        fixed_strength_threshold_value=0.0,
        threshold_sensitivity=0.0,
        signal_quality=3.0,
        update_rate=None,
        level_tracking_active=False,
    )


def get_medium_config() -> RefAppConfig:
    return RefAppConfig(
        median_filter_length=3,
        num_medians_to_average=1,
        start_m=0.050,
        end_m=6.0,
        max_profile=a121.Profile.PROFILE_5,
        num_frames_in_recorded_threshold=50,
        peaksorting_method=PeakSortingMethod.STRONGEST,
        reflector_shape=ReflectorShape.PLANAR,
        threshold_method=ThresholdMethod.CFAR,
        close_range_leakage_cancellation=False,
        fixed_threshold_value=100.0,
        fixed_strength_threshold_value=3.0,
        threshold_sensitivity=0.0,
        signal_quality=19.0,
        update_rate=None,
        level_tracking_active=False,
    )


def get_large_config() -> RefAppConfig:
    return RefAppConfig(
        median_filter_length=1,
        num_medians_to_average=1,
        start_m=0.10,
        end_m=15,
        max_profile=a121.Profile.PROFILE_5,
        num_frames_in_recorded_threshold=50,
        peaksorting_method=PeakSortingMethod.STRONGEST,
        reflector_shape=ReflectorShape.PLANAR,
        threshold_method=ThresholdMethod.CFAR,
        close_range_leakage_cancellation=False,
        fixed_threshold_value=100.0,
        fixed_strength_threshold_value=5.0,
        threshold_sensitivity=0.0,
        signal_quality=20.0,
        update_rate=None,
        level_tracking_active=False,
    )



# Copyright (c) Acconeer AB, 2023-2024
# All rights reserved

from __future__ import annotations

from typing import Optional, Tuple

import attrs
import numpy as np
import numpy.typing as npt
from scipy.signal import welch

from acconeer.exptool import a121
from acconeer.exptool.a121.algo import PERCEIVED_WAVELENGTH, AlgoProcessorConfigBase, ProcessorBase


@attrs.mutable(kw_only=True)
class ProcessorConfig(AlgoProcessorConfigBase):
    threshold: Optional[float] = attrs.field(default=10.0)

    num_segments: int = attrs.field(default=3)

    def _collect_validation_results(
        self, config: a121.SessionConfig
    ) -> list[a121.ValidationResult]:
        validation_results: list[a121.ValidationResult] = []

        return validation_results


@attrs.frozen(kw_only=True)
class ProcessorExtraResult:
    """
    Contains information for visualization in ET.
    """

    velocities: npt.NDArray[np.float64]
    psd: npt.NDArray[np.float64]
    est_peaks: npt.NDArray[np.float64]
    actual_thresholds: npt.NDArray[np.float64]


@attrs.frozen(kw_only=True)
class ProcessorResult:
    extra_result: ProcessorExtraResult
    speed_per_depth: npt.NDArray[np.float64]

    @property
    def max_speed(self) -> np.float64:
        return max(np.min(self.speed_per_depth), np.max(self.speed_per_depth), key=np.abs)


class Processor(ProcessorBase[ProcessorResult]):
    def __init__(
        self,
        *,
        sensor_config: a121.SensorConfig,
        metadata: a121.Metadata,
        processor_config: ProcessorConfig,
    ) -> None:
        self.threshold = processor_config.threshold
        self.num_points = sensor_config.num_points
        self.num_segments = processor_config.num_segments
        self.segment_length = sensor_config.sweeps_per_frame // self.num_segments

        if self.segment_length < 2:
            self.segment_length = 2

        assert sensor_config.sweep_rate is not None
        assert sensor_config.continuous_sweep_mode is not None

        self.sweep_rate = sensor_config.sweep_rate

    def get_welch(
        self, sweep: npt.NDArray[np.complex128]
    ) -> Tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
        freqs, psd = welch(
            sweep,
            fs=self.sweep_rate,
            window="hann",
            nperseg=self.segment_length,
            noverlap=0,
            average="mean",
            axis=0,
            return_onesided=False,
        )
        psd = np.fft.fftshift(psd, axes=0)
        freqs = np.fft.fftshift(freqs, axes=0)
        return freqs, psd

    def interpolate_peak(self, freqs: npt.NDArray[np.float64], peak_ind: int) -> float:
        # we assume indices to be -1,0,1 and take a inverse based on that.

        freqs = np.squeeze(freqs)
        if peak_ind == len(freqs) - 1 or peak_ind == 0:
            return peak_ind
        y1 = freqs[peak_ind - 1]
        y2 = freqs[peak_ind]
        y3 = freqs[peak_ind + 1]
        ys = np.array([y1, y2, y3])
        # Inverse of [[1,-1,1],[0,0,1],[1,1,1]]
        # fast_mat = np.array([[0.5, -1.0, 0.5], [-0.5, 0.0, 0.5], [0.0,1.0,0.0]])
        fast_mat = np.array([[0.5, -1.0, 0.5], [-0.5, 0.0, 0.5]])
        coeffs = np.dot(fast_mat, ys)
        max_ind = -coeffs[1] / (2 * coeffs[0])
        return float(peak_ind + max_ind)

    def interpolate_linear(self, speeds: npt.NDArray[np.float64], peak: float) -> float:
        p1 = int(np.floor(peak))
        p2 = int(np.ceil(peak))
        diff = speeds[p2] - speeds[p1]
        return float(speeds[p1] + diff * (peak - p1))

    def process(self, result: a121.Result) -> ProcessorResult:
        freqs, psd = self.get_welch(result.frame)

        speeds = freqs * PERCEIVED_WAVELENGTH

        peak_inds = np.argmax(psd, axis=0)

        speed_estimates = np.zeros(self.num_points)
        real_peaks = np.zeros(self.num_points)
        actual_thresholds = np.zeros(self.num_points)

        for i, peak_ind in enumerate(peak_inds):
            median = np.median(psd[:, i])
            norm_vals = psd[:, i] / median
            actual_thresholds[i] = median * self.threshold

            if norm_vals[peak_ind] > self.threshold:
                real_peak = self.interpolate_peak(norm_vals, peak_ind)
                real_speed = self.interpolate_linear(speeds, real_peak)
                real_peaks[i] = real_peak
                speed_estimates[i] = real_speed
            else:
                speed_estimates[i] = 0.0
                real_peaks[i] = 0.0

        extra_result = ProcessorExtraResult(
            velocities=speeds,
            psd=psd,
            est_peaks=real_peaks,
            actual_thresholds=actual_thresholds,
        )

        processor_result = ProcessorResult(
            extra_result=extra_result, speed_per_depth=speed_estimates
        )

        return processor_result



# Copyright (c) Acconeer AB, 2022-2025
# All rights reserved

from __future__ import annotations

import copy
import enum
from typing import Any, Optional, Tuple, Union

import numpy as np
import numpy.typing as npt
from scipy.signal import butter, filtfilt

from acconeer.exptool import a121
from acconeer.exptool.a121.algo import AlgoParamEnum


ENVELOPE_FWHM_M = {
    a121.Profile.PROFILE_1: 0.04,
    a121.Profile.PROFILE_2: 0.07,
    a121.Profile.PROFILE_3: 0.14,
    a121.Profile.PROFILE_4: 0.19,
    a121.Profile.PROFILE_5: 0.32,
}
APPROX_BASE_STEP_LENGTH_M = 2.5e-3
# Parameter of signal temperature model.
SIGNAL_TEMPERATURE_MODEL_PARAMETER = {
    a121.Profile.PROFILE_1: 67.0,
    a121.Profile.PROFILE_2: 85.0,
    a121.Profile.PROFILE_3: 86.0,
    a121.Profile.PROFILE_4: 99.0,
    a121.Profile.PROFILE_5: 104.0,
}
# Largest measurable distance per PRF.
MAX_MEASURABLE_DIST_M = {prf: prf.mmd for prf in set(a121.PRF)}
# Slope and interception of linear noise temperature model.
DEVIATION_TEMPERATURE_MODEL_PARAMETER = [-0.00275, 0.98536]

SPEED_OF_LIGHT = 299792458
RADIO_FREQUENCY = 60.5e9
WAVELENGTH = SPEED_OF_LIGHT / RADIO_FREQUENCY
PERCEIVED_WAVELENGTH = WAVELENGTH / 2

MEAN_ABS_DEV_OUTLIER_TH = 5

RLG_PER_HWAAS_MAP = {
    a121.Profile.PROFILE_1: 11.3,
    a121.Profile.PROFILE_2: 13.7,
    a121.Profile.PROFILE_3: 19.0,
    a121.Profile.PROFILE_4: 20.5,
    a121.Profile.PROFILE_5: 21.6,
}

SPARSE_IQ_PPC = 24


class ReflectorShape(AlgoParamEnum):
    """Reflector shape.

    ``GENERIC`` Reflectors of any shape.
    ``PLANAR`` Planar shaped reflectors facing the radar, for example water surfaces.
    """

    GENERIC = 4
    PLANAR = 2

    @property
    def exponent(self) -> float:
        return float(self.value)


class PeakSortingMethod(AlgoParamEnum):
    """Peak sorting methods.
    ``CLOSEST`` sort according to distance.
    ``STRONGEST`` sort according to strongest reflector."""

    CLOSEST = enum.auto()
    STRONGEST = enum.auto()


def get_distance_offset(
    peak_location: Optional[float], profile: a121.Profile, temperature: Optional[Union[float, int]]
) -> float:
    """
    Returns the distance offset in meters based on a loopback measurement.
    """
    if peak_location is None:
        return 0.0

    if temperature is not None:
        if temperature < -25:
            OFFSET_COMPENSATION_COEFFS = {
                a121.Profile.PROFILE_1: [0.81796372, 0.01561084],
                a121.Profile.PROFILE_2: [0.60363902, 0.01639924],
                a121.Profile.PROFILE_3: [0.76755397, 0.0119841],
                a121.Profile.PROFILE_4: [0.91310206, 0.01822777],
                a121.Profile.PROFILE_5: [1.05453651, 0.0045255],
            }
        elif temperature < 45:
            OFFSET_COMPENSATION_COEFFS = {
                a121.Profile.PROFILE_1: [0.91880272, 0.01782158],
                a121.Profile.PROFILE_2: [0.59457893, 0.0192985],
                a121.Profile.PROFILE_3: [0.85039166, 0.01478081],
                a121.Profile.PROFILE_4: [1.00348084, 0.02306631],
                a121.Profile.PROFILE_5: [0.98183729, 0.00794888],
            }
        elif temperature < 80:
            OFFSET_COMPENSATION_COEFFS = {
                a121.Profile.PROFILE_1: [0.54980311, 0.00836143],
                a121.Profile.PROFILE_2: [0.49726231, 0.01961523],
                a121.Profile.PROFILE_3: [0.53138948, 0.00784231],
                a121.Profile.PROFILE_4: [0.44605765, 0.01097863],
                a121.Profile.PROFILE_5: [0.51716572, 0.00060455],
            }
        else:
            OFFSET_COMPENSATION_COEFFS = {
                a121.Profile.PROFILE_1: [0.54980311, 0.00836143],
                a121.Profile.PROFILE_2: [0.49726231, 0.01961523],
                a121.Profile.PROFILE_3: [0.53138948, 0.00784231],
                a121.Profile.PROFILE_4: [0.44605765, 0.01097863],
                a121.Profile.PROFILE_5: [0.51716572, 0.00060455],
            }
    else:
        OFFSET_COMPENSATION_COEFFS = {
            a121.Profile.PROFILE_1: [0.91880272, 0.01782158],
            a121.Profile.PROFILE_2: [0.59457893, 0.0192985],
            a121.Profile.PROFILE_3: [0.85039166, 0.01478081],
            a121.Profile.PROFILE_4: [1.00348084, 0.02306631],
            a121.Profile.PROFILE_5: [0.98183729, 0.00794888],
        }

    p = OFFSET_COMPENSATION_COEFFS[profile]
    return p[0] * peak_location + p[1]


def _subsweep_distances(
    subsweep: a121.SubsweepConfig, metadata: a121.Metadata
) -> npt.NDArray[np.float64]:
    points = np.arange(subsweep.num_points) * subsweep.step_length + subsweep.start_point
    distances_m = np.array(points, dtype=float) * metadata.base_step_length_m
    return distances_m


def get_distances_m(
    config: Union[a121.SensorConfig, a121.SubsweepConfig], metadata: a121.Metadata
) -> npt.NDArray[np.float64]:
    """
    Returns an array of all distances measured by the config.
    The distances are returned in the same order as found in the result frame
    (:py:meth:`acconeer.exptool.a121.Result.frame`)
    """

    if isinstance(config, a121.SensorConfig):
        # Go through all subsweeps
        all_distances = []
        for subsweep in config.subsweeps:
            distances_m = _subsweep_distances(subsweep, metadata)
            all_distances += list(distances_m)
        ret = np.array(all_distances)
    else:
        # We are already a subsweep
        ret = _subsweep_distances(config, metadata)

    return ret


def get_approx_fft_vels(
    metadata: a121.Metadata, config: a121.SensorConfig
) -> Tuple[npt.NDArray[np.float64], float]:
    if config.sweep_rate is not None:
        sweep_rate = config.sweep_rate
    else:
        sweep_rate = metadata.max_sweep_rate

    spf = config.sweeps_per_frame
    f_res = 1 / spf
    freqs = np.fft.fftshift(np.fft.fftfreq(spf))
    f_to_v = 2.5e-3 * sweep_rate
    return freqs * f_to_v, f_res * f_to_v


def interpolate_peaks(
    abs_sweep: npt.NDArray[np.float64],
    peak_idxs: list[int],
    start_point: int,
    step_length: int,
    step_length_m: float,
) -> Tuple[list[float], list[float]]:
    """Quadratic interpolation around a peak using the amplitudes of the peak and its two
    neighboring points.

    Derivation:
    https://math.stackexchange.com/questions/680646/get-polynomial-function-from-3-points

    :param abs_sweep: Absolute value of mean sweep.
    :param peak_idxs: List containing indexes of identified peaks.
    :param start_point: Start point.
    :param step_length: Step length in points.
    :param step_length_m: Step length in meters.
    """
    estimated_distances = []
    estimated_amplitudes = []
    for peak_idx in peak_idxs:
        x = np.arange(peak_idx - 1, peak_idx + 2, 1)
        y = abs_sweep[peak_idx - 1 : peak_idx + 2]
        a = (x[0] * (y[2] - y[1]) + x[1] * (y[0] - y[2]) + x[2] * (y[1] - y[0])) / (
            (x[0] - x[1]) * (x[0] - x[2]) * (x[1] - x[2])
        )
        b = (y[1] - y[0]) / (x[1] - x[0]) - a * (x[0] + x[1])
        c = y[0] - a * x[0] ** 2 - b * x[0]
        peak_loc = -b / (2 * a)
        estimated_distances.append((start_point + peak_loc * step_length) * step_length_m)
        estimated_amplitudes.append(a * peak_loc**2 + b * peak_loc + c)
    return estimated_distances, estimated_amplitudes


def calculate_loopback_peak_location(result: a121.Result, config: a121.SensorConfig) -> float:
    """
    Calculate the distance tot peak of the loopback using interpolation.
    """

    (B, A) = get_distance_filter_coeffs(config.profile, config.step_length, narrow_filter=True)
    sweep = np.squeeze(result.frame, axis=0)
    abs_sweep = np.abs(filtfilt(B, A, sweep))
    peak_idx = [int(np.argmax(abs_sweep))]

    (estimated_dist, _) = interpolate_peaks(
        abs_sweep=abs_sweep,
        peak_idxs=peak_idx,
        start_point=config.start_point,
        step_length=config.step_length,
        step_length_m=APPROX_BASE_STEP_LENGTH_M,
    )

    return estimated_dist[0]


def find_peaks(
    abs_sweep: npt.NDArray[np.float64], threshold: npt.NDArray[np.float64]
) -> list[int]:
    """Identifies peaks above threshold.

    A peak is defined as a point with greater value than its two neighboring points and all
    three points are above the threshold.

    :param abs_sweep: Absolute value of mean sweep.
    :param threshold: Array of values, defining the threshold throughout the sweep.
    """
    if threshold is None:
        raise ValueError
    found_peaks = []
    d = 1
    N = len(abs_sweep)
    while d < (N - 1):
        if np.isnan(threshold[d - 1]):
            d += 1
            continue
        if np.isnan(threshold[d + 1]):
            break
        if abs_sweep[d] <= threshold[d]:
            d += 2
            continue
        if abs_sweep[d - 1] <= threshold[d - 1]:
            d += 1
            continue
        if abs_sweep[d - 1] >= abs_sweep[d]:
            d += 1
            continue
        d_upper = d + 1
        while True:
            if (d_upper) >= (N - 1):
                break
            if np.isnan(threshold[d_upper]):
                break
            if abs_sweep[d_upper] <= threshold[d_upper]:
                break
            if abs_sweep[d_upper] > abs_sweep[d]:
                break
            elif abs_sweep[d_upper] < abs_sweep[d]:
                found_peaks.append(int(np.argmax(abs_sweep[d:d_upper]) + d))
                break
            else:
                d_upper += 1
        d = d_upper
    return found_peaks


def get_temperature_adjustment_factors(
    reference_temperature: float, current_temperature: float, profile: a121.Profile
) -> Tuple[float, float]:
    """Calculate temperature compensation for mean sweep and background noise(tx off) standard
    deviation.

    The signal adjustment models how the amplitude level fluctuates with temperature.
    If the same object is measured against while the temperature changes,
    the amplitude level should be multiplied with this factor.
    The temperature difference should be calculated by subtracting
    the (calibration) recorded temperature from the measured.

    Example of usage:
    reference_temperature (recorded temperature during calibration)
    reference_amplitude (recorded amplitude during calibration)

    measurement_temperature (temperature at measurement time)

    signal_adjustment_factor, deviation_adjustment_factor =
    get_temperature_adjustment_factors(reference_temperature, measurement_temperature, profile)

    reference_amplitude_new = reference_amplitude * signal_adjustment_factor

    The reference_amplitude_new is an approximation of what the calibrated amplitude
    would be at the new temperature.

    E.g. When the temperature falls 60 degrees, the amplitude (roughly) doubles.
    This yields a signal_adjustment_factor of (about) 2.

    The signal adjustment model is follows 2 ** (temperature_diff / model_parameter), where
    model_parameter reflects the temperature difference relative the reference temperature,
    required for the amplitude to double/halve.

    The deviation_adjustment_factor works the same way, but is applied to a measurement
    taken with the Tx off. So if instead of measurement_amplitude, we have a measurement of tx_off.
    The procedure for calculating this is to take the same configuration as
    the application will use, but turning off the Tx.
    This calibration value is multiplied with the deviation_adjustment_factor.
    """

    temperature_diff = current_temperature - reference_temperature

    signal_adjustment_factor = 2 ** (
        (temperature_diff / SIGNAL_TEMPERATURE_MODEL_PARAMETER[profile]) * -1
    )
    deviation_adjustment_factor = (
        DEVIATION_TEMPERATURE_MODEL_PARAMETER[0] * temperature_diff
        + DEVIATION_TEMPERATURE_MODEL_PARAMETER[1]
    )
    return (signal_adjustment_factor, deviation_adjustment_factor)


def get_distance_filter_coeffs(
    profile: a121.Profile, step_length: int, narrow_filter: bool = False
) -> Any:
    """Calculates the IIR coefficients corresponding to a matched filter, based on the profile and
    the step length.

    Narrow filter increase the bandwidth of the filter, yielding a less smeared envelope after
    filtering.
    """
    NARROW_FILTER_MULTIPLIER = 2.0
    wnc = APPROX_BASE_STEP_LENGTH_M * step_length / (ENVELOPE_FWHM_M[profile])
    if narrow_filter:
        wnc *= NARROW_FILTER_MULTIPLIER
    return butter(N=1, Wn=wnc)


def get_distance_filter_edge_margin(profile: a121.Profile, step_length: int) -> int:
    """Calculates the number of points required for filter initialization when performing
    distance filtering, using the filter coefficients supplied by the function
    get_distance_filter_coeffs.
    """
    return int(_safe_ceil(ENVELOPE_FWHM_M[profile] / (APPROX_BASE_STEP_LENGTH_M * step_length)))


def double_buffering_frame_filter(
    _frame: npt.NDArray[Any],
) -> Optional[npt.NDArray[np.complex128]]:
    """
    Detects and removes outliers in data that appear when the double buffering mode is enabled,
    and returns the filtered frame.

    Outliers are detected along the sweep dimension using the second order difference. For
    reliable outlier detection, the filter is applied only when there are 32 or more sweeps per frame.

    The disturbance caused by enabling the double buffering mode can appear in multiple sweeps
    but, according to observations, is limited to a maximum of two consecutive sweeps. Therefore, the
    function removes outliers by interpolating between the sample before and the sample two positions
    ahead.

    The function does not correct disturbances that may appear in the initial or final sweeps.
    """

    (n_s, n_d) = _frame.shape
    min_num_sweeps = 32

    if n_s < min_num_sweeps:
        return None

    frame_real = _frame["real"]
    frame_imag = _frame["imag"]

    # Second order difference along sweeps
    frame_diff_real = np.zeros((n_s, n_d), dtype=np.int16)
    frame_diff_imag = np.zeros((n_s, n_d), dtype=np.int16)
    frame_diff_real[1:-1, :] = np.diff(frame_real, axis=0, n=2)
    frame_diff_imag[1:-1, :] = np.diff(frame_imag, axis=0, n=2)

    # Estimating magnitude using: abs(real) + abs(imag)
    frame_diff_abs = np.abs(frame_diff_real) + np.abs(frame_diff_imag)

    # Mean absolute deviation
    frame_diff_mad = np.sum(frame_diff_abs, axis=0) // (n_s - 2)

    # Detect outliers
    threshold = MEAN_ABS_DEV_OUTLIER_TH * frame_diff_mad
    outliers = frame_diff_abs > threshold

    # Perform filtering at each distance to remove outliers
    filtered_frame_real = frame_real.copy()
    filtered_frame_imag = frame_imag.copy()
    for d in range(n_d):
        if np.any(outliers[:, d]):
            args = np.where(outliers[:, d])[0]
            for idx in args:
                if idx <= 1:
                    # Median filtering for the first two and the last two sweeps
                    filtered_frame_real[idx, d] = np.median(filtered_frame_real[idx : idx + 4, d])
                    filtered_frame_imag[idx, d] = np.median(filtered_frame_imag[idx : idx + 4, d])
                elif idx >= n_s - 2:
                    filtered_frame_real[idx, d] = np.median(filtered_frame_real[idx - 3 : idx, d])
                    filtered_frame_imag[idx, d] = np.median(filtered_frame_imag[idx - 3 : idx, d])
                else:
                    # Interpolation for the remaining sweeps
                    filtered_frame_real[idx, d] = int(
                        (
                            2 * filtered_frame_real[max(idx - 1, 0), d]
                            + filtered_frame_real[min(idx + 2, n_s - 1), d]
                        )
                        / 3
                    )
                    filtered_frame_imag[idx, d] = int(
                        (
                            2 * filtered_frame_imag[max(idx - 1, 0), d]
                            + filtered_frame_imag[min(idx + 2, n_s - 1), d]
                        )
                        / 3
                    )

    filtered_frame = np.empty((n_s, n_d), dtype=np.complex128)
    filtered_frame.real = filtered_frame_real
    filtered_frame.imag = filtered_frame_imag

    return filtered_frame


def select_prf_m(distance_m: float, profile: a121.Profile) -> a121.PRF:
    """Calculates the highest possible PRF for the given breakpoint.

    :param distance_m: Distance in meters
    :param profile: Profile.
    """
    max_meas_dist_m = copy.copy(MAX_MEASURABLE_DIST_M)

    if a121.PRF.PRF_19_5_MHz in max_meas_dist_m and profile != a121.Profile.PROFILE_1:
        del max_meas_dist_m[a121.PRF.PRF_19_5_MHz]

    viable_prfs = [prf for prf, max_dist_m in max_meas_dist_m.items() if distance_m < max_dist_m]
    return sorted(viable_prfs, key=lambda prf: prf.frequency)[-1]


def select_prf(breakpoint: int, profile: a121.Profile) -> a121.PRF:
    """Calculates the highest possible PRF for the given breakpoint.

    :param breakpoint: Distance in the unit of points.
    :param profile: Profile.
    """
    return select_prf_m(breakpoint * APPROX_BASE_STEP_LENGTH_M, profile)


def get_max_profile_without_direct_leakage(start_m: float) -> a121.Profile:
    """
    Returns the highest possible profile such that
    the direct leakage for that profile doesn't include the start_m distance.
    If all direct leakages is inside the distance, return profile_1.
    """
    envelope_fwhm_m_reversed = dict(reversed(list(ENVELOPE_FWHM_M.items())))
    for profile, fwhm in envelope_fwhm_m_reversed.items():
        if fwhm * 2.0 <= start_m:
            return profile
    return a121.Profile.PROFILE_1


def get_max_step_length(profile: a121.Profile) -> int:
    """
    Calculate biggest possible step length based on the fwhm of the set profile.
    Achieve detection on the complete range with minimum number of sampling points.
    """
    fwhm_p = ENVELOPE_FWHM_M[profile] / APPROX_BASE_STEP_LENGTH_M
    if fwhm_p < SPARSE_IQ_PPC:
        step_length = SPARSE_IQ_PPC // int(np.ceil(SPARSE_IQ_PPC / fwhm_p))
    else:
        step_length = int((fwhm_p // SPARSE_IQ_PPC) * SPARSE_IQ_PPC)
    return step_length


def estimate_frame_rate(client: a121.Client, session_config: a121.SessionConfig) -> float:
    """
    Performs a measurement of the actual frame rate obtained by the configuration.
    This is hardware dependent. Hence the solution using a measurement.

    If a recorder is attached to the client,
    this call will result in a new session being run and recorded!
    """

    delta_times = np.full(2, np.nan)

    client.setup_session(session_config)
    client.start_session()

    for i in range(4):
        result = client.get_next()
        assert isinstance(result, a121.Result)

        if i < 2:
            # Ignore first read, it is sometimes inaccurate
            last_time = result.tick_time
            continue

        time = result.tick_time
        delta = time - last_time
        last_time = time
        delta_times = np.roll(delta_times, -1)
        delta_times[-1] = delta

    client.stop_session()

    return float(1.0 / np.nanmean(delta_times))


def exponential_smoothing_coefficient(fs: float, time_constant: float) -> float:
    """Calculate the exponential smoothing coefficient.

    Typical usage:

    y = y * coeff + x * (1 - coeff)

    :param fs: Sampling frequency.
    :param time_constant: Time constant.
    """
    dt = 1 / fs
    return float(np.exp(-dt / time_constant))


def _safe_ceil(x: float) -> float:
    """Perform safe ceil.

    Implementation of ceil function, compatible with float representation in C.
    """
    return float(f"{x:.16g}")


def calc_processing_gain(profile: a121.Profile, step_length: int) -> float:
    """
    Approximates the processing gain of the matched filter.
    """
    envelope_base_length_m = ENVELOPE_FWHM_M[profile] * 2  # approx envelope width
    num_points_in_envelope = (
        int(envelope_base_length_m / (step_length * APPROX_BASE_STEP_LENGTH_M)) + 2
    )
    mid_point = num_points_in_envelope // 2
    pulse = np.concatenate(
        (
            np.linspace(0, 1, mid_point),
            np.linspace(1, 0, num_points_in_envelope - mid_point),
        )
    )
    return float(np.sum(pulse**2))


def _convert_multiple_amplitudes_to_strengths(
    amplitudes: list[float],
    distances: list[float],
    subsweeps: list[a121.SubsweepConfig],
    bg_noise_std: list[float],
    reflector_shape: ReflectorShape,
) -> list[float]:
    # Determine subsweep breakpoints in meters.
    start_points = [subsweep.start_point for subsweep in subsweeps]
    bpts_m = np.array(start_points) * APPROX_BASE_STEP_LENGTH_M

    strengths = []
    # Loop over amplitude/distance pairs.
    for amplitude, distance in zip(amplitudes, distances):
        # For the current distance, get corresponding sensor parameters and background noise.
        subsweep_idx = np.sum(bpts_m < distance) - 1
        subsweep_config = subsweeps[subsweep_idx]
        sigma = bg_noise_std[subsweep_idx]

        # Calculate strengths.
        strengths.append(
            _convert_amplitude_to_strength(
                subsweep_config, amplitude, distance, sigma, reflector_shape
            )
        )

    return strengths


def _convert_amplitude_to_strength(
    subsweep_config: a121.SubsweepConfig,
    amplitude: float,
    distance: float,
    bg_noise_std: float,
    reflector_shape: ReflectorShape = ReflectorShape.GENERIC,
) -> float:
    processing_gain_db = 10 * np.log10(
        calc_processing_gain(subsweep_config.profile, subsweep_config.step_length)
    )
    s_db = 20 * np.log10(amplitude)
    n_db = 20 * np.log10(bg_noise_std)
    r_db = reflector_shape.exponent * 10 * np.log10(distance)
    rlg_db = RLG_PER_HWAAS_MAP[subsweep_config.profile] + 10 * np.log10(subsweep_config.hwaas)

    return float(s_db - n_db - rlg_db + r_db - processing_gain_db)



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

from typing import Any, Callable, Dict, Iterator, List, Optional, Type, Union

from PySide6.QtCore import QTimer
from PySide6.QtWidgets import QApplication

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool._bs_thread import BSThread, BSThreadDiedException
from acconeer.exptool.a121 import algo
from acconeer.exptool.a121.algo._base import InputT, ProcessorConfigT, ResultT
from acconeer.exptool.a121.algo._plugins.processor import SetupMessage
from acconeer.exptool.app.new.backend import PlotMessage
from acconeer.exptool.app.new.pluginbase import PlotPluginBase

from ._null_app_model import NullAppModel


_ProcessorGetter = Callable[
    [
        a121.SessionConfig,
        ProcessorConfigT,
        Union[a121.Metadata, List[Dict[int, a121.Metadata]]],
    ],
    algo.GenericProcessorBase[InputT, ResultT],
]

_SessionConfigGetter = Callable[
    [int],  # sensor id
    a121.SessionConfig,
]


def processor_main(
    *,
    processor_getter: _ProcessorGetter[ProcessorConfigT, InputT, ResultT],
    plot_plugin: Type[PlotPluginBase],
    session_config_getter: _SessionConfigGetter,
    processor_config_getter: Callable[[], ProcessorConfigT],
    _blinkstick_updater_cls: Optional[Any] = None,
) -> None:
    parser = a121.ExampleArgumentParser()
    parser.add_argument("--sensor", type=int, default=1)
    args = parser.parse_args()
    et.utils.config_logging(args)

    client = a121.Client.open(**a121.get_client_args(args))

    session_config = session_config_getter(args.sensor)

    metadata = client.setup_session(session_config)

    processor_config = processor_config_getter()
    processor = processor_getter(session_config, processor_config, metadata)

    qapp = QApplication()
    pg.setConfigOption("background", "w")
    pg.setConfigOption("foreground", "k")
    pg.setConfigOptions(antialias=True)

    plot_plugin_widget = plot_plugin(NullAppModel())
    plot_plugin_widget.handle_message(
        SetupMessage(
            session_config=session_config,
            metadata=metadata,
            processor_config=processor_config,
        )
    )

    if _blinkstick_updater_cls is None:
        bs_process = None
    else:
        bs_updater = _blinkstick_updater_cls()
        bs_process = BSThread(bs_updater)
        bs_process.start()

    client.start_session()

    interrupt_handler = et.utils.ExampleInterruptHandler()
    print("Press Ctrl-C to end session")

    loop = get_loop(client, processor, plot_plugin_widget, bs_process)

    timer = QTimer()
    timer.timeout.connect(lambda: qapp.quit() if not next(loop) else None)
    timer.timeout.connect(lambda: qapp.quit() if interrupt_handler.got_signal else None)
    timer.start()

    plot_plugin_widget.show()
    qapp.exec()

    print("Disconnecting...")

    if bs_process is not None:
        bs_process.close()

    client.close()


def get_loop(
    client: a121.Client,
    processor: algo.GenericProcessorBase[InputT, ResultT],
    plot_plugin_widget: PlotPluginBase,
    blinkstick_process: Optional[BSThread],
) -> Iterator[bool]:
    while True:
        result = client.get_next()

        processor_result = processor.process(result)  # type: ignore[arg-type]

        plot_plugin_widget.handle_message(PlotMessage(result=processor_result))
        plot_plugin_widget.draw()

        if blinkstick_process is not None:
            try:
                blinkstick_process.put_data(processor_result)
            except BSThreadDiedException:
                break

        yield True

    yield False



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

from __future__ import annotations

import os
from typing import Tuple, Union

import h5py


PathOrH5File = Union[str, "os.PathLike[str]", h5py.File]


def h5_file_factory(path_or_file: PathOrH5File, h5_file_mode: str) -> Tuple[h5py.File, bool]:
    """Constructs a `h5py.File` given an argument of type `PathOrFile`.

    If the argument already is a `h5py.File`, it's simply passed through.

    :returns: Tuple of (<h5py.File>, <whether the h5py.File was constructed in this function>)
    :raises: TypeError if `path_or_file` is an unsupported type.
    """
    if isinstance(path_or_file, h5py.File):
        return path_or_file, False
    elif isinstance(path_or_file, (os.PathLike, str)):
        path = path_or_file
        return h5py.File(path, mode=h5_file_mode), True
    else:
        msg = f"`path_or_file` was of unexpected type: {type(path_or_file)}"
        raise TypeError(msg)



# Copyright (c) Acconeer AB, 2025
# All rights reserved


from ._ex_app import CargoPresenceConfig, ContainerSize, ExAppConfig, UtilizationLevelConfig


def get_10_ft_container_config() -> ExAppConfig:
    return ExAppConfig(
        container_size=ContainerSize.CONTAINER_10_FT,
        activate_utilization_level=True,
        activate_presence=False,
    )


def get_20_ft_container_config() -> ExAppConfig:
    return ExAppConfig(
        container_size=ContainerSize.CONTAINER_20_FT,
        activate_utilization_level=True,
        activate_presence=False,
    )


def get_40_ft_container_config() -> ExAppConfig:
    return ExAppConfig(
        container_size=ContainerSize.CONTAINER_40_FT,
        activate_utilization_level=True,
        utilization_level_config=UtilizationLevelConfig(update_rate=3),
        activate_presence=False,
    )


def get_no_lens_config() -> ExAppConfig:
    return ExAppConfig(
        container_size=ContainerSize.CONTAINER_20_FT,
        activate_utilization_level=True,
        utilization_level_config=get_no_lens_utilization_level_config(),
        activate_presence=False,
        cargo_presence_config=get_no_lens_cargo_presence_config(),
    )


def get_no_lens_utilization_level_config() -> UtilizationLevelConfig:
    return UtilizationLevelConfig(
        update_rate=5,
        signal_quality=25,
        threshold_sensitivity=0.5,
    )


def get_no_lens_cargo_presence_config() -> CargoPresenceConfig:
    return CargoPresenceConfig(
        burst_rate=0.1,
        signal_quality=30,
        inter_detection_threshold=2,
        intra_detection_threshold=2,
    )



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from ._a121 import A121BackendPluginBase, A121ViewPluginBase
from ._processor_main import processor_main
from .processor import (
    ExtendedProcessorBackendPluginBase,
    ProcessorBackendPluginBase,
    ProcessorBackendPluginSharedState,
    ProcessorPluginPreset,
    ProcessorPluginSpec,
    ProcessorViewPluginBase,
    SetupMessage,
)



# Copyright (c) Acconeer AB, 2023-2025
# All rights reserved

from __future__ import annotations

import abc
import collections.abc
import json
import numbers
import typing as t

import attrs
import exceptiongroup as eg
import h5py
import typing_extensions as te

from . import core


_T = t.TypeVar("_T")
TypeLike = t.Union[
    type,
    t._GenericAlias,  # type: ignore[name-defined]
]


class LoadError(Exception):
    def __repr__(self) -> str:
        return str(self)


class SaveError(Exception):
    def __repr__(self) -> str:
        return str(self)


class SaveErrorGroup(SaveError, eg.ExceptionGroup[Exception]):
    pass


class LoadErrorGroup(LoadError, eg.ExceptionGroup[Exception]):
    pass


class MissingH5ObjectError(LoadError):
    @classmethod
    def create(cls, parent: h5py.Group, name: str) -> te.Self:
        return cls(
            f"Expected group {parent} to contain child {name!r}, "
            + f"but it only contains {list(parent.keys())}"
        )


class WrongH5ObjectError(LoadError):
    def __init__(
        self,
        parent: h5py.Group,
        child: t.Union[h5py.Dataset, h5py.Group],
        expected_type: type,
    ) -> None:
        super().__init__(f"Member {child} of {parent} was expected to be a {expected_type}")


class TypeMissmatchError(SaveError):
    @classmethod
    def wrong_type_encountered(cls, instance: t.Any, path: str, *expected_types: type) -> te.Self:
        if expected_types == ():
            msg = "Gotta specify at least one expected type"
            raise ValueError(msg)

        if len(expected_types) == 1:
            (expected_type,) = expected_types
            return cls(
                f"Expected type {expected_type} at {path!r}. Encountered an instance of type {type(instance)}: {instance}"
            )
        else:
            return cls(
                f"Expected one of types {expected_types} at {path!r}. Encountered an instance of type {type(instance)}: {instance}"
            )


class TypeMissmatchErrorGroup(TypeMissmatchError, eg.ExceptionGroup[TypeMissmatchError]):
    """Exception group for TypeMissmatchErrors"""


class JsonPresentable(te.Protocol):
    def to_json(self) -> str: ...

    @classmethod
    def from_json(cls, json_str: str) -> te.Self: ...


@attrs.frozen
class Persistor(abc.ABC):
    PRIORITY: t.ClassVar[int] = 0

    parent_group: h5py.Group
    name: str
    type_tree: core.Node

    @classmethod
    def from_json_presentable(cls, presentable_type: t.Type[JsonPresentable]) -> t.Type[Persistor]:
        """Generates a Persistor class for the specified JsonPresentable"""

        def _is_applicable(cls: t.Type[Persistor], __type: core.TypeLike) -> bool:
            return core.is_subclass(__type, presentable_type)

        def _save(self: Persistor, data: t.Any) -> None:
            if not isinstance(data, presentable_type):
                raise TypeMissmatchError.wrong_type_encountered(data, "?", presentable_type)

            self.create_own_dataset(data=data.to_json())

        def _load(self: Persistor) -> JsonPresentable:
            dataset_contents = self.dataset[()]
            cls.assert_not_empty(dataset_contents)
            try:
                return presentable_type.from_json(bytes.decode(dataset_contents))
            except json.JSONDecodeError:
                raise LoadError

        return type(
            f"GeneratedPersistor_JsonPresentable_{presentable_type.__name__}_{hash(presentable_type)}",
            (Persistor,),
            {
                "is_applicable": classmethod(_is_applicable),
                "_save": _save,
                "_load": _load,
            },
        )

    @staticmethod
    def assert_not_empty(dataset_contents: t.Any) -> None:
        """Asserts that the dataset contents is not empty"""
        if isinstance(dataset_contents, h5py.Empty):
            raise core.LoadError

    @staticmethod
    def assert_empty(dataset_contents: t.Any) -> None:
        """Asserts that the dataset contents is not empty"""
        if not isinstance(dataset_contents, h5py.Empty):
            raise core.LoadError

    @property
    def group(self) -> h5py.Group:
        obj = self.parent_group.get(self.name, default=None)

        if isinstance(obj, h5py.Group):
            return obj
        elif obj is None:
            raise MissingH5ObjectError.create(self.parent_group, self.name)
        else:
            raise WrongH5ObjectError(self.parent_group, obj, expected_type=h5py.Group)

    @property
    def dataset(self) -> h5py.Dataset:
        obj = self.parent_group.get(self.name, default=None)

        if isinstance(obj, h5py.Dataset):
            return self.parent_group[self.name]
        elif obj is None:
            raise MissingH5ObjectError.create(self.parent_group, self.name)
        else:
            raise WrongH5ObjectError(self.parent_group, obj, expected_type=h5py.Dataset)

    def require_own_group(self) -> h5py.Group:
        group = self.parent_group.require_group(self.name)
        group.attrs["persistor"] = f"{type(self).__name__}"
        return group

    def create_own_dataset(
        self,
        data: t.Any,
        *,
        dtype: t.Any = None,
        shape: t.Optional[tuple[int, ...]] = None,
    ) -> h5py.Dataset:
        if self.name == "./":
            msg = "Datasets cannot be the root entry. Try wrapping the object or save a different representation."
            msg += f"\n\nAttempted to save the following to './': {data!r}"
            raise SaveError(msg)
        dataset = self.parent_group.create_dataset(self.name, data=data, dtype=dtype, shape=shape)
        dataset.attrs["persistor"] = f"{type(self).__name__}"
        return dataset

    @t.final
    def save(self, data: t.Any) -> None:
        try:
            self._save(data)
        except SaveError as se:
            raise se
        except Exception as e:
            raise SaveError from e

    @t.final
    def load(self) -> t.Any:
        try:
            res = self._load()
        except LoadError as le:
            raise le
        except Exception as e:
            raise LoadError from e
        else:
            return res

    @classmethod
    @abc.abstractmethod
    def is_applicable(cls, __type: core.TypeLike) -> bool:
        pass

    @abc.abstractmethod
    def _save(self, data: t.Any) -> None:
        pass

    @abc.abstractmethod
    def _load(self) -> t.Any:
        pass


@attrs.frozen
class Node:
    """Basic tree node implementation."""

    data: type
    children: dict[str, Node]


def is_class(__type: TypeLike) -> te.TypeGuard[type]:
    """Return True if '__type' is a class"""
    return not core.is_ndarray(__type) and isinstance(__type, type)


def is_subclass(__type: TypeLike, superclass: t.Union[type, t.Tuple[type, ...]]) -> bool:
    """Return True if '__type' is a subclass of superclass"""
    cannot_be_generic: t.Iterable[type]
    if isinstance(superclass, type):
        cannot_be_generic = [superclass]
    else:
        cannot_be_generic = superclass

    for sc in cannot_be_generic:
        if is_generic(sc):
            msg = f"A type passed as the 'superclass' argument cannot be a generic class. It was {sc}"
            raise ValueError(msg)

    try:
        if is_generic(__type):
            (origin_t, *_) = unwrap_generic(__type)
            return is_subclass(origin_t, superclass)
        else:
            return is_class(__type) and issubclass(__type, superclass)
    except Exception as e:
        msg = f"Could not determine if {__type} is a subclass of {superclass}"
        raise TypeError(msg) from e


def is_generic(__type: TypeLike) -> bool:
    """Return True if '__type' is a generic type"""
    return te.get_origin(__type) is not None and te.get_args(__type) != ()


def is_optional(__type: TypeLike) -> bool:
    """Return True if '__type' is an optional type"""
    origin, *type_args = unwrap_generic(__type)

    return origin is t.Union and type(None) in type_args and len(type_args) > 1


def optional_args(__type: TypeLike) -> t.Tuple[TypeLike, ...]:
    _, *type_args = unwrap_generic(__type)

    return tuple(ta for ta in type_args if ta is not type(None))


def optional_arg(__type: TypeLike) -> TypeLike:
    (arg,) = optional_args(__type)
    return arg


def unwrap_generic(__type: TypeLike) -> t.Sequence[TypeLike]:
    """
    'Unwraps' a generic returning its origin and type args.

    For example:
        List[int] |-> (list, int)
        Dict[int, Enum] |-> (dict, int, Enum)
    """
    return (te.get_origin(__type), *te.get_args(__type))


def is_ndarray(__type: TypeLike) -> bool:
    # Extra check for numpy arrays (annotated with 'numpy.typing.NDArray').
    #   In versions <3.9/3.10, numpy has their own GenericAlias that does not play
    # well with the introspection functions from 'typing'.
    return getattr(__type, "__name__", None) == "ndarray"


def get_class_type_hints(__type: type) -> dict[str, TypeLike]:
    try:
        return t.get_type_hints(__type)
    except TypeError as error:
        raise TypeError(
            f"{__type} is annotated with built-ins (list, dict, etc.). "
            + "Use 'typing' counterpart instead (typing.List, typing.Dict, etc.)"
        ) from error


def sequence_index(index: t.Union[str, int] = "X") -> str:
    return f"sequence_index_{index}"


def union_index(index: int) -> str:
    return f"union_index_{index}"


DICT_KEY_TYPE = "dict_key_type"
DICT_VALUE_TYPE = "dict_value_type"


def create_type_tree(__type: TypeLike, attr_path: str = "") -> Node:
    """Creates a tree of Nodes that acts as a schema for serializing
    instances of `__type`.

    :param __type: The type to create a type tree for
    :param attr_path: Accumulated attribute path. Used for enhancing error messages.
    :returns: The root of the type tree.
    """
    if isinstance(__type, t.TypeVar):
        msg = "Cannot resolve type variables of generic classes"
        raise NotImplementedError(msg)

    if is_ndarray(__type):
        hints = {}
    elif is_generic(__type):
        origin, *type_args = unwrap_generic(__type)

        if origin in [list, t.Sequence, collections.abc.Sequence]:
            (type_arg,) = type_args
            hints = {sequence_index(): type_arg}

        elif origin in [dict, t.Mapping, collections.abc.Mapping]:
            (key_type, value_type) = type_args
            hints = {
                DICT_KEY_TYPE: key_type,
                DICT_VALUE_TYPE: value_type,
            }

        elif origin is tuple:
            if ... in type_args:
                msg = f"Unsupported Tuple with ellipsis (...) at {attr_path!r}"
                raise NotImplementedError(msg)

            hints = {sequence_index(i): arg for i, arg in enumerate(type_args)}

        elif origin is t.Union:
            hints = {union_index(i): arg for i, arg in enumerate(type_args)}

        else:
            msg = f"Unexpected generic type at {attr_path!r}: '{__type}' (origin={origin}, type_args={type_args})"
            raise ValueError(msg)
    elif is_class(__type):
        hints = get_class_type_hints(__type)
    else:
        msg = f"Unexpected type at {attr_path!r}: '{__type}' of type {type(__type)}"
        raise RuntimeError(msg)

    try:
        return Node(__type, {name: create_type_tree(hint) for name, hint in hints.items()})
    except RecursionError:
        msg = f"{__type} is recursive. Recursively defined classes are not supported"
        raise TypeError(msg)


def sanitize_instance(instance: t.Any, type_tree: Node, attr_path: str = "") -> None:
    """Asserts that the instance conforms to its type and type annotations with instance-checks

    :param instance: The instance to sanitize
    :param type_tree: The type tree (schema) to use for sanitation
    :param attr_path: Accumulated attribute path. Used for enhancing error messages.
    """

    current_type = type_tree.data

    if is_ndarray(current_type):
        return
    elif is_generic(current_type):
        origin, *type_args = unwrap_generic(current_type)

        if origin is t.Union:
            children = type_tree.children.items()
            errors = []
            for _, child in children:
                try:
                    sanitize_instance(instance, child, attr_path=f"{attr_path}")
                except TypeMissmatchError as tme:
                    errors.append(tme)
                    continue
                else:
                    break
            else:
                msg = "No Union members were sanitary"
                raise TypeMissmatchErrorGroup(msg, errors)

            return

        if not isinstance(instance, origin):
            raise TypeMissmatchError.wrong_type_encountered(instance, attr_path, origin)

        if origin in [list, t.Sequence, collections.abc.Sequence]:
            (child,) = type_tree.children.values()

            for idx, value in enumerate(instance):
                sanitize_instance(value, child, attr_path=f"{attr_path}[{idx}]")

            return

        if origin in [dict, t.Mapping, collections.abc.Mapping]:
            (key_tree, value_tree) = type_tree.children.values()

            for key, value in instance.items():
                sanitize_instance(key, key_tree, attr_path=f"{attr_path}.keys()[{key}]")
                sanitize_instance(value, value_tree, attr_path=f"{attr_path}[{key}]")

            return

        if origin is tuple:
            if ... in type_args:
                msg = "Tuple with ellipsis (...) is not supported"
                raise NotImplementedError(msg)

            children = type_tree.children.items()

            assert len(children) == len(instance)
            for value, (child_name, child) in zip(instance, children):
                sanitize_instance(value, child, attr_path=f"{attr_path}[{child_name}]")

            return

        else:
            msg = f"Unexpected generic type '{current_type}' (origin={origin}, type_args={type_args})"
            raise ValueError(msg)

        msg = "Fell through instance sanitization"
        raise RuntimeError(msg)
    elif is_subclass(current_type, numbers.Number) and isinstance(instance, numbers.Number):
        return
    elif is_class(current_type):
        if isinstance(instance, current_type):
            for name, tree in type_tree.children.items():
                child = getattr(instance, name)
                sanitize_instance(child, tree, attr_path=f"{attr_path}.{name}")
        else:
            raise TypeMissmatchError.wrong_type_encountered(instance, attr_path, current_type)



# Copyright (c) Acconeer AB, 2023-2025
# All rights reserved

from acconeer.exptool import a121

from ._detector import DetectorConfig


"""
Some of the config parameters will not be applied,
since they are overridden by automatic_subsweeps.
They are still left in to allow for a simple disabling.
"""


def get_short_range_config() -> DetectorConfig:
    return DetectorConfig(
        start_m=0.06,
        end_m=1.0,
        frame_rate=10,
        automatic_subsweeps=True,
        signal_quality=30.0,
        sweeps_per_frame=16,
        hwaas=16,
        inter_frame_idle_state=a121.IdleState.DEEP_SLEEP,
        intra_enable=True,
        intra_detection_threshold=1.4,
        intra_frame_time_const=0.15,
        intra_output_time_const=0.3,
        inter_enable=True,
        inter_detection_threshold=1,
        inter_frame_slow_cutoff=0.2,
        inter_frame_fast_cutoff=5,
        inter_frame_deviation_time_const=0.5,
        inter_output_time_const=2,
        inter_frame_presence_timeout=3,
    )


def get_medium_range_config() -> DetectorConfig:
    return DetectorConfig(
        start_m=0.3,
        end_m=2.5,
        frame_rate=12,
        automatic_subsweeps=True,
        signal_quality=20.0,
        sweeps_per_frame=16,
        hwaas=32,
        inter_frame_idle_state=a121.IdleState.DEEP_SLEEP,
        intra_enable=True,
        intra_detection_threshold=1.3,
        intra_frame_time_const=0.15,
        intra_output_time_const=0.3,
        inter_enable=True,
        inter_detection_threshold=1,
        inter_frame_slow_cutoff=0.2,
        inter_frame_fast_cutoff=6,
        inter_frame_deviation_time_const=0.5,
        inter_output_time_const=2,
        inter_frame_presence_timeout=3,
    )


def get_long_range_config() -> DetectorConfig:
    return DetectorConfig(
        start_m=5,
        end_m=7.5,
        frame_rate=12,
        automatic_subsweeps=True,
        signal_quality=10.0,
        sweeps_per_frame=16,
        hwaas=128,
        inter_frame_idle_state=a121.IdleState.DEEP_SLEEP,
        intra_enable=True,
        intra_detection_threshold=1.2,
        intra_frame_time_const=0.15,
        intra_output_time_const=0.3,
        inter_enable=True,
        inter_detection_threshold=0.8,
        inter_frame_slow_cutoff=0.2,
        inter_frame_fast_cutoff=6,
        inter_frame_deviation_time_const=0.5,
        inter_output_time_const=2,
        inter_frame_presence_timeout=3,
    )


def get_low_power_config() -> DetectorConfig:
    return DetectorConfig(
        start_m=0.38,
        end_m=0.67,
        frame_rate=0.7,
        automatic_subsweeps=False,
        signal_quality=20.0,
        sweeps_per_frame=8,
        hwaas=8,
        profile=a121.Profile.PROFILE_5,
        inter_frame_idle_state=a121.IdleState.DEEP_SLEEP,
        intra_enable=True,
        intra_detection_threshold=1.7,
        intra_frame_time_const=0.3,
        intra_output_time_const=0.3,
        inter_enable=True,
        inter_detection_threshold=1.2,
        inter_frame_slow_cutoff=0.2,
        inter_frame_fast_cutoff=5,
        inter_frame_deviation_time_const=0.5,
        inter_output_time_const=0.5,
        inter_frame_presence_timeout=2,
    )
